/*
 *
 * ftsmooth.c
 *
 * Anti-aliasing renderer interface (body).
 *
 * Copyright 2000-2015 by
 * David Turner, Robert Wilhelm, and Werner Lemberg.
 *
 * Copyright 2019-2021 MicroEJ Corp. This file has been modified by MicroEJ Corp.
 *
 * This file is part of the FreeType project, and may only be used,
 * modified, and distributed under the terms of the FreeType project
 * license, LICENSE.TXT.  By continuing to use, modify, or distribute
 * this file you indicate that you have read the license and
 * understand and accept it fully.
 */

#define FT_MAKE_OPTION_SINGLE_OBJECT

#include "ft2build.h"

#include FT_FREETYPE_H
#include FT_INTERNAL_OBJECTS_H

#include "ftsmerrs.h"

#include "ft2build.h"
#include FT_INTERNAL_DEBUG_H
#include FT_OUTLINE_H
#include "ftspic.h"

#include "intern/ftvglite_intern.h"

#undef  FT_COMPONENT
#define FT_COMPONENT  trace_smooth

#ifndef FT_MEM_SET
#define FT_MEM_SET( d, s, c )  ft_memset( d, s, c )
#endif

#ifndef FT_MEM_ZERO
#define FT_MEM_ZERO( dest, count )  FT_MEM_SET( dest, 0, count )
#endif

#include "trace_vglite.h"

/*
 * @brief Logs an init gpu start event
 *
 * @param[in] operation: The VGLite operation sent to the GPU
 */
#define FT_VGLITE_TRACE_INIT_GPU_START(operation) \
        TRACE_PLATFORM_START_U32( \
                VGLITE, \
                VGLITE_TRACE_INIT_GPU, \
                VGLITE_TRACE_OP_ ## operation \
                        );

/*
 * @brief Logs an init gpu end event
 */
#define FT_VGLITE_TRACE_INIT_GPU_END() \
        TRACE_PLATFORM_END_VOID( \
                VGLITE, \
                VGLITE_TRACE_INIT_GPU \
                        );

/*
 * @brief INIT_GPU event identifier
 */
#define VGLITE_TRACE_INIT_GPU 1

/*
 * @brief VGLite operation DRAW (draw path) identifier
 * This event is expected to be used to trace a call to vg_lite_draw
 */
#define VGLITE_TRACE_OP_DRAW 2

/*
 * @brief VGLite operation DRAW_GRAD (draw path with gradient) identifier
 * This event is expected to be used to trace a call to vg_lite_draw_grad
 */
#define VGLITE_TRACE_OP_DRAW_GRAD 3


typedef struct  vglite_TWorker_ {
	FT_Outline  outline;

	vg_lite_buffer_t* vglite_buffer_addr;
	vg_lite_matrix_t* vglite_matrix;
	vg_lite_path_t vglite_path;
	uint32_t vglite_path_pool_size;
	vg_lite_fill_t vg_lite_fill_rule;
	vg_lite_blend_t vg_lite_blend;
	vg_lite_color_t vg_lite_color;
	vg_lite_linear_gradient_t* vg_lite_gradient;

	FT_Glyph_Metrics    *metrics;
} vglite_TWorker, *vglite_PWorker;

static vglite_TWorker  _worker;

typedef struct vglite_TRaster_ {
	void* memory;
} vglite_TRaster, *vglite_PRaster;

static vglite_TRaster _raster;

static FT_Error vglite_extend_path(int extra_length){
    FT_Error error = FT_ERR( Invalid_Argument );
    FT_Memory memory = _raster.memory;
    int32_t new_path_length;

    new_path_length = _worker.vglite_path.path_length + extra_length;

    if ( new_path_length <= _worker.vglite_path_pool_size ) {
        error = FT_ERR( Ok );
    } else {
        do {
            if ( FT_REALLOC( _worker.vglite_path.path, _worker.vglite_path_pool_size, _worker.vglite_path_pool_size + FT_VGLITE_PATH_POOL_CHUNK ) ) {
                error = FT_ERR( Out_Of_Memory );
                break;
            }

            _worker.vglite_path_pool_size += FT_VGLITE_PATH_POOL_CHUNK;
        } while ( new_path_length > _worker.vglite_path_pool_size );
        error = FT_ERR( Ok );
    }

    if ( FT_ERR( Ok ) == error ) {
        _worker.vglite_path.path_length = new_path_length;
    }

    return error;
}

static FT_Error vglite_get_path_slot(void **slot, int extra_length) {
    FT_Error    error;
    int32_t     _prev_path_length;

    *slot = NULL;

    _prev_path_length = _worker.vglite_path.path_length;

    error = vglite_extend_path(extra_length);
    if ( FT_ERR( Ok ) == error ) {
        *slot = &(((uint8_t *) _worker.vglite_path.path) [_prev_path_length]);;
    }

    return error;
}

static int
vglite_move_to( const FT_Vector*    to,
                vglite_PWorker      worker )
{
    FT_Error            error;
    path_move_to_16_t   *path_move_to;
    void                *path_slot;

    FT_TRACE7(("%s: \n", __func__));

    error = vglite_get_path_slot(&path_slot, sizeof(*path_move_to));
    if ( FT_ERR( Ok ) == error ) {
        FT_TRACE7(("move to %d, %d\n", to->x, to->y));
        path_move_to = path_slot;
        path_move_to->cmd = _VGLITE_CMD_MOVE_TO;
        path_move_to->x = to->x;
        path_move_to->y = -to->y;
    }

    return error;
}

static int
vglite_line_to( const FT_Vector*    to,
                vglite_PWorker      worker )
{
    FT_Error            error;
    path_line_to_16_t   *path_line_to;
    void                *path_slot;

    FT_TRACE7(("%s: \n", __func__));

    error = vglite_get_path_slot(&path_slot, sizeof(*path_line_to));
    if ( FT_ERR( Ok ) == error ) {
        FT_TRACE7(("line to %d, %d\n", to->x, to->y));
        path_line_to = path_slot;
        path_line_to->cmd = _VGLITE_CMD_LINE_TO;
        path_line_to->x = to->x;
        path_line_to->y = -to->y;
    }

    return error;
}


static int
vglite_conic_to(    const FT_Vector*    control,
                    const FT_Vector*    to,
                    vglite_PWorker      worker )
{
    FT_Error            error;
    path_quad_to_16_t   *path_quad_to;
    void                *path_slot;

    FT_TRACE7(("%s: \n", __func__));

    error = vglite_get_path_slot(&path_slot, sizeof(*path_quad_to));
    if ( FT_ERR( Ok ) == error ) {
        FT_TRACE7(("quad to %d, %d, %d, %d\n", control->x, control->y, to->x, to->y));
        path_quad_to = path_slot;
        path_quad_to->cmd = _VGLITE_CMD_QUAD_TO;
        path_quad_to->cx = control->x;
        path_quad_to->cy = -control->y;
        path_quad_to->x = to->x;
        path_quad_to->y = -to->y;
    }

    return error;
}


static int
vglite_cubic_to(    const FT_Vector*    control1,
                    const FT_Vector*    control2,
                    const FT_Vector*    to,
                    vglite_PWorker      worker )
{
    FT_Error            error;
    path_cubic_to_16_t  *path_cubic_to;
    void                *path_slot;

    FT_TRACE7(("%s: \n", __func__));

    error = vglite_get_path_slot(&path_slot, sizeof(*path_cubic_to));
    if ( FT_ERR( Ok ) == error ) {
        FT_TRACE7(("cubic to %d, %d, %d, %d, %d, %d\n", control1->x, control1->y, control2->x, control2->y, to->x, to->y));
        path_cubic_to = path_slot;
        path_cubic_to->cmd = _VGLITE_CMD_CUBIC_TO;
        path_cubic_to->cx1 = control1->x;
        path_cubic_to->cy1 = -control1->y;
        path_cubic_to->cx2 = control2->x;
        path_cubic_to->cy2 = -control2->y;
        path_cubic_to->x = to->x;
        path_cubic_to->y = -to->y;
    }

    return error;
}

static int
vglite_end( vglite_PWorker  worker )
{
    FT_Error            error;
    path_end_16_t       *path_end;
    void                *path_slot;

    FT_TRACE7(("%s: \n", __func__));

    error = vglite_get_path_slot(&path_slot, sizeof(*path_end));
    if ( FT_ERR( Ok ) == error ) {
        FT_TRACE7(("end\n"));
        path_end = path_slot;
        path_end->cmd = _VGLITE_CMD_END;
    }

    return error;
}

FT_DEFINE_OUTLINE_FUNCS(func_interface,
    (FT_Outline_MoveTo_Func) vglite_move_to,
    (FT_Outline_LineTo_Func) vglite_line_to,
    (FT_Outline_ConicTo_Func)vglite_conic_to,
    (FT_Outline_CubicTo_Func)vglite_cubic_to,
    0,
    0
)

static int
vglite_convert_glyph( void )
{
    volatile int  error = 0;

    FT_TRACE7(("%s: \n", __func__));

    // Reinit path.
    _worker.vglite_path.path_length = 0;

    error = FT_Outline_Decompose( &_worker.outline, &func_interface, &_worker );

    FT_TRACE7(( "Send path to gpu\n" ));
    vglite_end( &_worker );
    _worker.vglite_path.path_changed = 1;

    FT_TRACE7(("width: %d\n", _worker.metrics->width));
    FT_TRACE7(("height: %d\n", _worker.metrics->height));
    FT_TRACE7(("horiBearingX: %d\n", _worker.metrics->horiBearingX));
    FT_TRACE7(("horiBearingY: %d\n", _worker.metrics->horiBearingY));

// FIXME the commented code break characters alignment on the baseline.
// I guess they were added to fix wrong positions.
//    int offX = _worker.metrics->horiBearingX;
//    int offY = _worker.metrics->horiBearingY - _worker.metrics->height;

//    vg_lite_translate(offX, offY, _worker.vglite_matrix);
    if(_worker.vg_lite_gradient == NULL){
		FT_VGLITE_TRACE_INIT_GPU_START(DRAW);
    	vg_lite_draw(_worker.vglite_buffer_addr, &_worker.vglite_path, _worker.vg_lite_fill_rule, _worker.vglite_matrix, _worker.vg_lite_blend, _worker.vg_lite_color);
		FT_VGLITE_TRACE_INIT_GPU_END();
    }
    else {
		FT_VGLITE_TRACE_INIT_GPU_START(DRAW_GRAD);
    	vg_lite_draw_gradient(_worker.vglite_buffer_addr, &_worker.vglite_path, _worker.vg_lite_fill_rule, _worker.vglite_matrix, _worker.vg_lite_gradient, _worker.vg_lite_blend);
		FT_VGLITE_TRACE_INIT_GPU_END();
    }
//    vg_lite_translate(0, offY, _worker.vglite_matrix);

    return error;
}

static int
vglite_render(  vglite_PRaster          raster,
                const FT_Raster_Params* params )
{
    const FT_Outline*  outline    = (const FT_Outline*)params->source;

    FT_TRACE7(("%s: \n", __func__));

    if ( &_raster != raster )
    {
        FT_TRACE7(("%s: internal error\n", __func__));
        return FT_THROW( Invalid_Argument );
    }

    _worker.outline        = *outline;

    return vglite_convert_glyph(  );
}


  /**** RASTER OBJECT CREATION: In stand-alone mode, we simply use *****/
  /****                         a static object.                   *****/

static void
vglite_reset(   FT_Raster  raster,
                char*      pool_base,
                long       pool_size )
{
    FT_TRACE7(("%s: \n", __func__));

    if ( (FT_Raster) &_raster != raster )
        FT_TRACE7(("%s: internal error\n", __func__));

    _worker.vglite_buffer_addr = NULL;
    _worker.vglite_matrix = NULL;
    _worker.vglite_path_pool_size = 0;

    memset(&_worker.vglite_path, 0, sizeof(_worker.vglite_path));

    _worker.vglite_path.bounding_box[0] = -4000; // Left
    _worker.vglite_path.bounding_box[1] = -4000; // Top
    _worker.vglite_path.bounding_box[2] = 4000; // Right
    _worker.vglite_path.bounding_box[3] = 4000; // Bottom
}

static int
vglite_new( FT_Memory   memory,
            FT_Raster*  araster )
{
    FT_TRACE7(("%s: \n", __func__));

    *araster = (FT_Raster) &_raster;
    _raster.memory = memory;

    vglite_reset( *araster, NULL, 0 );

    return FT_ERR( Ok );
}


static void
vglite_done( FT_Raster  raster )
{
    FT_TRACE7(("%s: \n", __func__));
    FT_Memory  memory = _raster.memory;

    if ( NULL != _worker.vglite_path.path ) {
        FT_FREE( _worker.vglite_path.path );
        _worker.vglite_path_pool_size = 0;
    }
}

FT_DEFINE_RASTER_FUNCS(ft_vglite_raster,
    FT_GLYPH_FORMAT_OUTLINE,

    (FT_Raster_New_Func)     vglite_new,
    (FT_Raster_Reset_Func)   vglite_reset,
    (FT_Raster_Set_Mode_Func)0,
    (FT_Raster_Render_Func)  vglite_render,
    (FT_Raster_Done_Func)    vglite_done
)

/* initialize renderer -- init its raster */
static FT_Error
ft_vglite_init( FT_Renderer  render )
{
    FT_Library  library = FT_MODULE_LIBRARY( render );

    FT_TRACE7(("%s: \n", __func__));

    render->clazz->raster_class->raster_reset( render->raster,
                                               library->raster_pool,
                                               library->raster_pool_size );

    return 0;
}


/* sets render-specific mode */
static FT_Error ft_vglite_set_mode(FT_Renderer render, FT_ULong  mode_tag, FT_Pointer data ){
	FT_Error ret = FT_ERR(Invalid_Argument);

	FT_TRACE7(("%s: \n", __func__));

    switch (mode_tag){
		case FT_PARAM_TAG_VGLITE_BUFFER_ADDR:{
			_worker.vglite_buffer_addr = data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_MATRIX:{
			_worker.vglite_matrix = data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_QUALITY:{
			_worker.vglite_path.quality = (vg_lite_quality_t) (uint32_t) data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_FORMAT:{
			_worker.vglite_path.format = (vg_lite_format_t) (uint32_t) data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_FILL_RULE:{
			_worker.vg_lite_fill_rule = (vg_lite_fill_t) (uint32_t) data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_BLEND:{
			_worker.vg_lite_blend = (vg_lite_blend_t) (uint32_t) data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_COLOR:{
			_worker.vg_lite_color = (uint32_t) data;
			ret = FT_ERR(Ok);
			break;
		}
		case FT_PARAM_TAG_VGLITE_GRADIENT:{
			_worker.vg_lite_gradient = data;
			ret = FT_ERR(Ok);
			break;
    	}
		default:{
			/* we simply pass it to the raster */
			ret = render->clazz->raster_class->raster_set_mode(render->raster, mode_tag, data);
			break;
		}
	}

	return ret;
}

/* transform a given glyph image */
static FT_Error
ft_vglite_transform(    FT_Renderer       render,
                        FT_GlyphSlot      slot,
                        const FT_Matrix*  matrix,
                        const FT_Vector*  delta )
{
    FT_Error  error = FT_Err_Ok;

    FT_TRACE7(("%s: \n", __func__));

    if ( slot->format != render->glyph_format )
    {
        error = FT_THROW( Invalid_Argument );
        goto Exit;
    }

    if ( matrix )
        FT_Outline_Transform( &slot->outline, matrix );

    if ( delta )
        FT_Outline_Translate( &slot->outline, delta->x, delta->y );

    Exit:
    return error;
}


/* return the glyph's control box */
static void
ft_vglite_get_cbox( FT_Renderer   render,
                    FT_GlyphSlot  slot,
                    FT_BBox*      cbox )
{
    FT_MEM_ZERO( cbox, sizeof ( *cbox ) );

    if ( slot->format == render->glyph_format )
        FT_Outline_Get_CBox( &slot->outline, cbox );
}

/* render a slot's glyph with GPU */
static FT_Error
ft_vglite_render(   FT_Renderer       render,
                    FT_GlyphSlot      slot,
                    FT_Render_Mode    mode,
                    const FT_Vector*  origin )
{
    FT_Error     error;
    FT_Outline*  outline = NULL;
    FT_BBox      cbox;
    FT_Pos       width, height;
//    FIXME These variables are used in commented code.
//    FT_Pos       x_shift = 0;
//    FT_Pos       y_shift = 0;

    FT_Raster_Params  params;

//    FT_Bool  have_translated_origin = FALSE;
//    FT_Bool  have_outline_shifted   = FALSE;

    FT_TRACE7(("%s: \n", __func__));

    _worker.metrics = &slot->metrics;

    /* check glyph image format */
    if ( slot->format != render->glyph_format )
    {
        error = FT_THROW( Invalid_Argument );

        goto Exit;
    }

    outline = &slot->outline;

//    FIXME Origin translation breaks string drawing, all the characters are
//    aligned on their descent line instead of their original baseline. For
//    instance, the lowest pixel of a 'g' or a 'q' will be at the same y
//    position as the lowest piwel of a 'a'.
//    /* translate the outline to the new origin if needed */
//    if ( origin )
//    {
//        FT_Outline_Translate( outline, origin->x, origin->y );
//        have_translated_origin = TRUE;
//    }

    /* compute the control box, and grid fit it */
    FT_Outline_Get_CBox( outline, &cbox );

    cbox.xMin = FT_PIX_FLOOR( cbox.xMin );
    cbox.yMin = FT_PIX_FLOOR( cbox.yMin );
    cbox.xMax = FT_PIX_CEIL( cbox.xMax );
    cbox.yMax = FT_PIX_CEIL( cbox.yMax );

    if ( cbox.xMin < 0 && cbox.xMax > FT_INT_MAX + cbox.xMin )
    {
        FT_ERROR((  "ft_vglite_render: glyph too large:"
                    " xMin = %d, xMax = %d\n",
                    cbox.xMin >> 6, cbox.xMax >> 6 ));
        error = FT_THROW( Raster_Overflow );
        goto Exit;
    }
    else
        width = ( cbox.xMax - cbox.xMin ) >> 6;

    if ( cbox.yMin < 0 && cbox.yMax > FT_INT_MAX + cbox.yMin )
    {
        FT_ERROR((  "ft_vglite_render: glyph too large:"
                    " yMin = %d, yMax = %d\n",
                    cbox.yMin >> 6, cbox.yMax >> 6 ));
        error = FT_THROW( Raster_Overflow );
        goto Exit;
    }
    else
        height = ( cbox.yMax - cbox.yMin ) >> 6;

//    FIXME see FIXME comment above.
//    /* allocate new one */
//    x_shift = (FT_Int) cbox.xMin;
//    y_shift = (FT_Int) cbox.yMin;

    /* Required check is (pitch * height < FT_ULONG_MAX),        */
    /* but we care realistic cases only.  Always pitch <= width. */
    if ( width > 0x7FFF || height > 0x7FFF )
    {
        FT_ERROR(( "ft_vglite_render: glyph too large: %u x %u\n",
                    width, height ));
        error = FT_THROW( Raster_Overflow );
        goto Exit;
    }

//    FIXME see FIXME comment above.
//    /* translate outline to render it into the bitmap */
//    FT_Outline_Translate( outline, -x_shift, -y_shift );
//    have_outline_shifted = TRUE;

    /* set up parameters */
    params.source = outline;
    params.flags  = FT_RASTER_FLAG_AA;

    /* render outline into bitmap */
    error = render->raster_render( render->raster, &params );

Exit:
//    FIXME see FIXME comment above.
//    if ( have_outline_shifted )
//        FT_Outline_Translate( outline, x_shift, y_shift );
//    if ( have_translated_origin )
//        FT_Outline_Translate( outline, -origin->x, -origin->y );

    return error;
}

FT_DEFINE_RENDERER( ft_vglite_renderer_class,

    FT_MODULE_RENDERER,
    sizeof ( FT_RendererRec ),

    "vglite_renderer",
    0x10000L,   // Version
    0x20000L,   // Requires

    0,    /* module specific interface */

    (FT_Module_Constructor)     ft_vglite_init,
    (FT_Module_Destructor)      0,
    (FT_Module_Requester)       0,

    FT_GLYPH_FORMAT_OUTLINE,

    (FT_Renderer_RenderFunc)    ft_vglite_render,
    (FT_Renderer_TransformFunc) ft_vglite_transform,
    (FT_Renderer_GetCBoxFunc)   ft_vglite_get_cbox,
    (FT_Renderer_SetModeFunc)   ft_vglite_set_mode,

    (FT_Raster_Funcs*)          &ft_vglite_raster
)

/* END */
