/*
 * Java
 *
 * Copyright 2020 MicroEJ Corp. All rights reserved.
 * Use of this source code is governed by a BSD-style license that can be found with this software.
 */
package com.microej.path.basic;

import static com.verisilicon.vivante.vg_lite_matrix.vg_lite_identity;

import com.microej.demo.watch.util.path.SinglePathGraphics;
import com.microej.path.PathHelper;
import com.microej.vivante.utils.Path;
import com.verisilicon.vivante.vg_lite_matrix;

import ej.drawing.ShapePainter.Cap;

/**
 * A path that represents a circle arc.
 */
public class CircleArc extends SinglePathGraphics {

	private float arcAngle;

	private final int diameter;

	private final int startAngle;

	private final int thickness;

	private final Cap cap;

	private int centerX;

	private int centerY;

	/**
	 * Creates a path that represents a circle arc.
	 *
	 * @param diameter
	 *            the diameter of the circular shape.
	 * @param thickness
	 *            the thickness of the shape.
	 * @param startAngle
	 *            the beginning angle of the arc, in degrees.
	 * @param color
	 *            the path color.
	 * @param cap
	 *            the cap to use for start and end of shape.
	 */
	/* package */ CircleArc(Path path, int diameter, int thickness, int startAngle, Cap cap) {
		super(path);
		this.diameter = diameter;
		this.thickness = thickness;
		this.startAngle = startAngle;
		this.cap = cap;
		this.arcAngle = -1;
		updatePath();
	}

	@Override
	protected void updateMatrices() {
		vg_lite_matrix matrix = getMatrix();
		assert matrix != null;
		vg_lite_identity(matrix);
		vg_lite_matrix.vg_lite_translate(this.centerX, this.centerY, matrix);
	}

	private void updatePath() {
		Path path = getPath();
		assert path != null;
		PathHelper.computeThickShapeEllipseArc(path.getVGLitePath(), this.diameter, this.diameter, this.thickness,
				this.startAngle, this.arcAngle, this.cap);
	}

	/**
	 * Sets the arc angle.
	 *
	 * <p>
	 * The arc starts at <code>startAngle</code> up to <code>arcAngle</code> degrees. <br>
	 * Angles are interpreted such that 0 degrees is at the 3 o'clock position. A positive <code>arcAngle</code> value
	 * indicates a counter-clockwise rotation whereas a negative value indicates a clockwise rotation.<br>
	 *
	 * @param arcAngle
	 *            the arc angle to set, in degrees.
	 */
	public void setArcAngle(float arcAngle) {
		this.arcAngle = arcAngle;
		updatePath();
		updateMatrices();
	}

	/**
	 * Gets the diameter.
	 *
	 * @return the diameter.
	 */
	protected int getDiameter() {
		return this.diameter;
	}

	/**
	 * Gets the thickness.
	 *
	 * @return the thickness.
	 */
	protected int getThickness() {
		return this.thickness;
	}

	/**
	 * Gets the start angle.
	 *
	 * @return the start angle.
	 */
	protected int getStartAngle() {
		return this.startAngle;
	}

	/**
	 * Sets the arc center point
	 *
	 * @param x
	 *            center point's X coordinate
	 * @param y
	 *            center point's Y coordinate
	 */
	protected void setCenter(int x, int y) {
		this.centerX = x;
		this.centerY = y;
	}

	/**
	 * Gets the arc center point's X coordinate
	 *
	 * @return the X coordinate
	 */
	protected int getCenterX() {
		return this.centerX;
	}

	/**
	 * Gets the arc center point's Y coordinate
	 *
	 * @return the Y coordinate
	 */
	protected int getCenterY() {
		return this.centerY;
	}

}
