/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.annotation.processor.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.tools.FileObject;

/**
 * A description is a model representing a {@link com.is2t.giml.description.Description} element.
 */
public class Description {

	private String _name; // NOSONAR
	private boolean _root; // NOSONAR
	private String _class; // NOSONAR
	private final List<Attribute> _attributes; // NOSONAR
	private final List<Child> _children; // NOSONAR
	private String _friendlyName; // NOSONAR
	private String _description; // NOSONAR
	private FileObject _icon16; // NOSONAR
	private FileObject _icon24; // NOSONAR
	private final List<CodeInterface> _codeInterfaces;

	/**
	 * Instantiate a new empty description.
	 */
	public Description() {
		this._attributes = new ArrayList<Attribute>();
		this._children = new ArrayList<Child>();
		this._codeInterfaces = new ArrayList<CodeInterface>();
	}

	/**
	 * Sets the name of this element.
	 *
	 * @param name
	 *            the name of this element.
	 */
	public void setName(String name) {
		this._name = name;
	}

	/**
	 * Sets if this element is a root, i.e. is a top-level element of descriptions.
	 *
	 * @param root
	 *            {@code true} if element is a top-level element, {@code false} otherwise.
	 */
	public void setRoot(boolean root) {
		this._root = root;
	}

	/**
	 * Sets the fully qualified name of this element type.
	 *
	 * @param clazz
	 *            the fully qualified name of this element type.
	 */
	public void setElementClassName(String clazz) {
		this._class = clazz;
	}

	/**
	 * Add an attribute to this element.
	 *
	 * @param attribute
	 *            the attribute to add.
	 */
	public void addAttribute(Attribute attribute) {
		this._attributes.add(attribute);
	}

	/**
	 * Add a child element to this element.
	 *
	 * @param child
	 *            the child element to add.
	 */
	public void addChild(Child child) {
		if (!this._children.contains(child)) {
			this._children.add(child);
		}
	}

	/**
	 * Gets the {@link CodeInterface} connected to this description.
	 *
	 * @return the {@link CodeInterface} connected to this description.
	 */
	public List<CodeInterface> getCodeInterfaces() {
		return _codeInterfaces;
	}

	/**
	 * Add a {@link CodeInterface} element to this element.
	 *
	 * @param ci
	 *            the {@link CodeInterface} element to add.
	 */
	public void addCodeInterface(CodeInterface ci) {
		this._codeInterfaces.add(ci);
	}

	/**
	 * Sets the friendly name.
	 *
	 * @param friendlyName
	 *            the name to set.
	 */
	public void setFriendlyName(String friendlyName) {
		this._friendlyName = friendlyName;
	}

	/**
	 * Sets the description of this element.
	 *
	 * @param description
	 *            the description to set.
	 */
	public void setDescription(String description) {
		this._description = description;
	}

	/**
	 * Sets the icon 16x16.
	 *
	 * @param icon
	 *            the icon to set.
	 */
	public void setIcon16(FileObject icon) {
		this._icon16 = icon;
	}

	/**
	 * Sets the icon 24x24.
	 *
	 * @param icon
	 *            the icon to set.
	 */
	public void setIcon24(FileObject icon) {
		this._icon24 = icon;
	}

	/**
	 * Gets the name of this element.
	 *
	 * @return the name of this element.
	 */
	public String getName() {
		return this._name;
	}

	/**
	 * Gets if this element is a top-level element.
	 *
	 * @return {@code true} if element is a top-level element, {@code false} otherwise.
	 */
	public boolean isRoot() {
		return this._root;
	}

	/**
	 * Gets the fully qualified name of this element type.
	 *
	 * @return the fully qualified name of this element type.
	 */
	public String getElementClassName() {
		return this._class;
	}

	/**
	 * Gets if this element has ordered children.
	 *
	 * @return {@code true} if this element has at least one ordered child, {@code false} otherwise.
	 *
	 * @see Child#isOrdered()
	 */
	public boolean hasOrderedChildren() {
		for (Child child : this._children) {
			if (child.isOrdered()) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Gets if this element has a default container in its children.
	 *
	 * @return {@code true} if this element has a default container in its children, {@code false} otherwise.
	 *
	 * @see Child#hasEmptyName()
	 */
	public boolean hasDefaultElement() {
		for (Child child : this._children) {
			if (child.hasEmptyName() && !child.isDisabled()) {
				return true;
			}
		}

		return false;
	}

	/**
	 * Gets a read-only list of this element attributes.
	 *
	 * @return a read-only list of this element attributes.
	 */
	public List<Attribute> getAttributes() {
		return Collections.unmodifiableList(this._attributes);
	}

	/**
	 * Gets a read-only list of this elements children.
	 *
	 * @return a read-only list of this elements children.
	 */
	public List<Child> getChildren() {
		List<Child> enabledChildren = new ArrayList<Child>();
		for (Child child : this._children) {
			if (!child.isDisabled()) {
				enabledChildren.add(child);
			}
		}
		return Collections.unmodifiableList(enabledChildren);
	}

	/**
	 * Gets a read-only list of this element children, without default elements.
	 *
	 * @return a read-only list of this element children, without default elements.
	 *
	 * @see #hasDefaultElement()
	 */
	public Set<Child> getNonDefaultChildren() {
		Set<Child> withoutDefaults = new HashSet<Child>();
		for (Child child : this._children) {
			if (!child.hasEmptyName() && !child.isDisabled()) {
				withoutDefaults.add(child);
			}
		}
		return Collections.unmodifiableSet(withoutDefaults);
	}

	/**
	 * Gets this element friendly name.
	 *
	 * @return the friendly name.
	 */
	public String getFriendlyName() {
		return this._friendlyName;
	}

	/**
	 * Gets this element description.
	 *
	 * @return the description.
	 */
	public String getDescription() {
		return this._description;
	}

	/**
	 * Gets the icon 16x16.
	 *
	 * @return the icon 16x16.
	 */
	public FileObject getIcon16() {
		return this._icon16;
	}

	/**
	 * Gets the icon 24x24.
	 *
	 * @return the icon 24x24.
	 */
	public FileObject getIcon24() {
		return this._icon24;
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append(this._name).append(':').append(this._class);
		return builder.toString();
	}

}
