/*
 * Java
 *
 * Copyright 2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.jsgenerator;

import java.util.ArrayList;
import java.util.List;

import com.is2t.giml.classpath.ClasspathModel;
import com.is2t.giml.classpath.ClasspathModel.InstanceMethod;
import com.is2t.giml.jsgenerator.JavaScriptSourceFunctionAnalyzer.AnalyzedFunction;
import com.is2t.giml.jsgenerator.global.GlobalObject;
import com.is2t.giml.jsgenerator.global.GlobalObjectFunction;
import com.is2t.giml.jsgenerator.global.GlobalObjectUserDefined;
import com.is2t.java.nodes.ClassDeclaration;
import com.is2t.java.nodes.Expression;
import com.is2t.java.nodes.ForStatement;
import com.is2t.java.nodes.IfStatement;
import com.is2t.java.nodes.LocalDeclaration;
import com.is2t.java.nodes.MethodDeclaration;
import com.is2t.java.nodes.Modifiers;
import com.is2t.java.nodes.ReturnStatement;
import com.is2t.java.nodes.ThrowStatement;
import com.is2t.java.nodes.TypeReference;
import com.is2t.java.nodes.expressions.ArrayAllocation;
import com.is2t.java.nodes.expressions.Assignment;
import com.is2t.java.nodes.expressions.BinaryExpression;
import com.is2t.java.nodes.expressions.BooleanLiteral;
import com.is2t.java.nodes.expressions.CastExpression;
import com.is2t.java.nodes.expressions.DoubleLiteral;
import com.is2t.java.nodes.expressions.FieldAccess;
import com.is2t.java.nodes.expressions.IntegerLiteral;
import com.is2t.java.nodes.expressions.Literal;
import com.is2t.java.nodes.expressions.LocalAccess;
import com.is2t.java.nodes.expressions.LongLiteral;
import com.is2t.java.nodes.expressions.MessageSend;
import com.is2t.java.nodes.expressions.NullLiteral;
import com.is2t.java.nodes.expressions.ObjectAllocation;
import com.is2t.java.nodes.expressions.StringLiteral;
import com.is2t.java.nodes.expressions.TernaryExpression;
import com.is2t.java.nodes.expressions.TypeReferenceExpression;

import jdk.nashorn.internal.ir.AccessNode;
import jdk.nashorn.internal.ir.BinaryNode;
import jdk.nashorn.internal.ir.Block;
import jdk.nashorn.internal.ir.CallNode;
import jdk.nashorn.internal.ir.ExpressionStatement;
import jdk.nashorn.internal.ir.ForNode;
import jdk.nashorn.internal.ir.FunctionNode;
import jdk.nashorn.internal.ir.IdentNode;
import jdk.nashorn.internal.ir.IfNode;
import jdk.nashorn.internal.ir.JoinPredecessorExpression;
import jdk.nashorn.internal.ir.LexicalContext;
import jdk.nashorn.internal.ir.LiteralNode;
import jdk.nashorn.internal.ir.Node;
import jdk.nashorn.internal.ir.ReturnNode;
import jdk.nashorn.internal.ir.Statement;
import jdk.nashorn.internal.ir.TernaryNode;
import jdk.nashorn.internal.ir.UnaryNode;
import jdk.nashorn.internal.ir.VarNode;
import jdk.nashorn.internal.ir.visitor.NodeVisitor;

/**
 *
 */
public class JavaScriptSourceToMicroEJSourceGenerator extends NodeVisitor<LexicalContext> {

	private final JavaScriptGenerator javaScriptGenerator;

	private final ArrayList<Expression> expressionStack;
	private Statement currentStatement;
	/**
	 * null if not building a function
	 */
	private AnalyzedFunction currentAnalyzedFunction;
	private com.is2t.java.nodes.Statement lastGeneratedJavaStatement;

	/**
	 * The fully qualified name of the class that play the role of "the global
	 * object"
	 */
	private String globalObjectMicroEJClassName;


	public final TypeReference TYPE_OBJECT;
	public final TypeReference TYPE_FLOAT;
	public final TypeReference TYPE_THIS_CLASS;

	/**
	 * List of instance method calls (name+number of parameters)
	 */
	private final List<InstanceMethodCall> instanceMethodsCalls;

	private final JavaScriptSourcePreGenerator jsSourcePreGenerator;

	public JavaScriptSourceToMicroEJSourceGenerator(JavaScriptGenerator javaScriptGenerator,
			JavaScriptSourcePreGenerator jsSourcePreGenerator) {
		super(new LexicalContext());
		this.javaScriptGenerator = javaScriptGenerator;
		this.jsSourcePreGenerator = jsSourcePreGenerator;
		expressionStack = new ArrayList<>();

		globalObjectMicroEJClassName = "";

		instanceMethodsCalls = new ArrayList<InstanceMethodCall>();
		TYPE_OBJECT = new TypeReference(javaFullyQualifiedName("Object"), 0);
		TYPE_FLOAT = new TypeReference(javaFullyQualifiedName("Float"), 0);
		TYPE_THIS_CLASS = new TypeReference(javaFullyQualifiedName(javaScriptGenerator.getGeneratedClassname()), 0);
	}

	/**
	 * API. Generate a statement
	 *
	 * @param statement
	 *            the JS statement
	 * @return the generated MicroEJ statement, or null if a function has been
	 *         generated
	 */
	public com.is2t.java.nodes.Statement generate(Statement statement) {
		this.currentStatement = statement;
		this.lastGeneratedJavaStatement = null;
		assert expressionStack.size() == 0;
		statement.accept(this);
		assert expressionStack.size() == 0;
		this.currentStatement = null;
		return lastGeneratedJavaStatement;
	}

	/**
	 * API Generate a JavaScript function.
	 *
	 * @see #getFunctions()
	 */
	public MethodDeclaration generate(FunctionNode node) {
		currentAnalyzedFunction = new JavaScriptSourceFunctionAnalyzer().generate(node);

		List<com.is2t.java.nodes.Statement> generatedStatementsVect = new ArrayList<>();
		// declare local variables
		for (String localName : currentAnalyzedFunction.getLocals()) {
			generatedStatementsVect
			.add(new LocalDeclaration(TYPE_OBJECT, javaName(localName), new Modifiers(), getUndefinedObject()));
		}

		List<Statement> statements = node.getBody().getStatements();
		for (Statement statement : statements) {
			com.is2t.java.nodes.Statement s = generate(statement);
			if (s != null) {
				generatedStatementsVect.add(s);
			}
		}

		// implicit return - TODO CCN Analysis
		generateImplicitReturn(generatedStatementsVect);

		com.is2t.java.nodes.Statement[] generatedStatements = generatedStatementsVect.toArray(new com.is2t.java.nodes.Statement[generatedStatementsVect.size()]);

		MethodDeclaration md = new MethodDeclaration(new ClassDeclaration(new char[0][]),
				new Modifiers(Modifiers.AccPublic | Modifiers.AccStatic),
				javaName(node.getName()), generateParameters(node.getParameters()), TYPE_OBJECT,
				generatedStatements);

		currentAnalyzedFunction = null;
		return md;
	}

	/**
	 * Generate function parameters
	 *
	 * @param parameters
	 * @return a list of {@link LocalDeclaration} declaring parameters (all
	 *         considered as Objects)
	 */
	private LocalDeclaration[] generateParameters(List<IdentNode> parameters) {
		ArrayList<LocalDeclaration> paramsVect = new ArrayList<>();
		for(IdentNode parameter : parameters) {
			paramsVect.add(new LocalDeclaration(TYPE_OBJECT, javaName(parameter.getName()), null));
		}
		return paramsVect.toArray(new LocalDeclaration[paramsVect.size()]);
	}

	@Override
	public Node leaveExpressionStatement(ExpressionStatement expressionStatement) {
		lastGeneratedJavaStatement = popExpression();
		return expressionStatement;
	}

	@Override
	public boolean enterLiteralNode(final LiteralNode<?> literalNode) {
		if (literalNode.isNull()) {
			pushExpression(new NullLiteral());
		} else if (literalNode.isNumeric()) {
			Object value = literalNode.getValue();
			if (value instanceof Integer) {
				Integer valueI = (Integer) value;
				pushExpression(newLiteral("Integer", new IntegerLiteral(valueI.intValue())));
			}
			else if (value instanceof Long) {
				Long valueL = (Long) value;
				pushExpression(newLiteral("Long", new LongLiteral(valueL.longValue())));
			} else if (value instanceof Double) {
				Double valueD = (Double) value;
				pushExpression(newLiteral("Double", new DoubleLiteral(valueD.doubleValue())));
			} else {
				throw new UnsupportedJavaScriptSyntax(literalNode);
			}
		} else if (literalNode.isString()) {
			pushExpression(new StringLiteral(literalNode.getString()));
		} else if (literalNode.getType().isBoolean()) {
			pushExpression(newBooleanLiteral(literalNode.getBoolean()));
		} else {
			throw new UnsupportedJavaScriptSyntax(literalNode);
		}
		return true;
	}

	private Expression newBooleanLiteral(boolean value) {
		return newLiteral("Boolean", new BooleanLiteral(value));
	}

	private Expression newLiteral(String javaLangTypeName, Expression arg) {
		return new ObjectAllocation(getJavaLangTypeReference(javaLangTypeName), new Expression[] { arg });
	}

	private TypeReference getJavaLangTypeReference(String simpleTypeName) {
		return new TypeReference(javaFullyQualifiedName("java.lang." + simpleTypeName), 0);
	}

	@Override
	public final Node leaveUnaryNode(final UnaryNode unaryNode) {
		Expression expr = popExpression();
		Expression newExpr;
		boolean isLiteral = expr instanceof Literal;
		switch (unaryNode.tokenType()) {
		// case ADD: {
		// if (isLiteral) {
		// // avoid a generated '+'
		// newExpr = expr;
		// } else {
		// newExpr = new UnaryExpression(UnaryExpression.Plus, expr);
		// }
		// break;
		// }
		// case BIT_NOT: {
		// newExpr = new UnaryExpression(UnaryExpression.BitwiseComplement, expr);
		// break;
		// }
		// case SUB:{
		// newExpr = null;
		// if (isLiteral) {
		// if (expr instanceof IntegerLiteral) {
		// // positive integer => negative integer
		// newExpr = new IntegerLiteral(-((IntegerLiteral) expr).value);
		// } else if (expr instanceof LongLiteral) {
		// // special case for Integer.MIN_VALUE
		// long value = ((LongLiteral) expr).value;
		// if (value == ((long) Integer.MAX_VALUE + 1)) {
		// newExpr = new IntegerLiteral(Integer.MIN_VALUE);
		// }
		// } else if (expr instanceof DoubleLiteral) {
		// // special case for Long.MIN_VALUE
		// double value = ((DoubleLiteral) expr).value;
		// if (value == 9.223372036854775808e18d) {
		// newExpr = new LongLiteral(Long.MIN_VALUE);
		// }
		// }
		// }
		//
		// if (newExpr == null) {
		// // default case
		// newExpr = new UnaryExpression(UnaryExpression.Minus, expr);
		// }
		// break;
		// }
		// case DECPREFIX: {
		// newExpr = new UnaryExpression(UnaryExpression.PreDecr, expr);
		// break;
		// }
		// case DECPOSTFIX: {
		// newExpr = new UnaryExpression(UnaryExpression.PostDecr, expr);
		// break;
		// }
		// case INCPREFIX: {
		// newExpr = new UnaryExpression(UnaryExpression.PreIncr, expr);
		// break;
		// }
		// case INCPOSTFIX:{
		// newExpr = new UnaryExpression(UnaryExpression.PostIncr, expr);
		// break;
		// }
		// case NOT: {
		// newExpr = new UnaryExpression(UnaryExpression.Not, expr);
		// break;
		// }
		case TYPEOF:
		case VOID:
		case DELETE:
		case NEW:{
			throw new UnsupportedJavaScriptSyntax(unaryNode);
		}
		default:
			throw new AssertionError();
		}

		// pushExpression(newExpr);
		// return unaryNode;
	}

	@Override
	public Node leaveIdentNode(IdentNode identNode) {
		Expression expr;
		expr = getVariableAccess(identNode);
		pushExpression(expr);
		return identNode;
	}

	@Override
	public boolean enterCallNode(CallNode callNode) {
		// extract function name and receiver
		jdk.nashorn.internal.ir.Expression function = callNode.getFunction();
		String functionName;
		jdk.nashorn.internal.ir.Expression receiver;
		if (function instanceof IdentNode) {
			functionName = ((IdentNode) function).getName();
			receiver = null;
		} else if (function instanceof AccessNode) {
			AccessNode accessNode = (AccessNode) function;
			functionName = accessNode.getProperty();
			receiver = accessNode.getBase();
		} else {
			throw new UnsupportedJavaScriptSyntax(callNode);
		}

		if (expandBuiltinNode(callNode, functionName, receiver)) {
			return false;
		}

		List<jdk.nashorn.internal.ir.Expression> args = callNode.getArgs();
		ArrayList<Expression> javaArgsVect = new ArrayList<>();
		for(jdk.nashorn.internal.ir.Expression arg : args) {
			arg.accept(this);
			javaArgsVect.add(popExpression());
		}

		Expression expr;
		if (receiver == null) {
			// Global object function
			MessageSend ms = new MessageSend(javaName(functionName),
					javaFullyQualifiedName(getGlobalObjectMicroEJClassName()), null, javaExpressionArray(javaArgsVect));
			expr = ms;
		}
		else {
			// generate the receiver
			generateReceiver(receiver);
			Expression javaReceiverExpression = popExpression();
			MessageSend ms;
			if (javaReceiverExpression instanceof TypeReferenceExpression) {
				// static call
				TypeReference type = ((TypeReferenceExpression) javaReceiverExpression).type;
				ms = new MessageSend(javaName(functionName), type, javaReceiverExpression,
						javaExpressionArray(javaArgsVect));
			} else {
				// instance call
				int nbParameters = args.size();

				// insert as first argument
				javaArgsVect.add(0, javaReceiverExpression);
				InstanceMethodCall call = new InstanceMethodCall(functionName, nbParameters);
				addInstanceMethodCall(call);
				ms = new MessageSend(getGeneratedMethodName(call), new char[0][], null,
						javaExpressionArray(javaArgsVect));
			}
			expr = ms;
		}

		pushExpression(expr);

		return false; // this node is responsible to visit sub-nodes
	}

	/**
	 * @see #convertToTargetType(Node, Expression, String)
	 */
	private Expression convertToBooleanType(Node source, Expression src) {
		return convertToTargetType(source, src, "B");
	}

	/**
	 * Convert an expression (valued to an Object) to the appropriate Java
	 * expression that will produce the value of the target type. <br>
	 * Rules:
	 * <li>String => toString()</li>
	 * <li>basetype => bind_[BT]Value0(Object)</li>
	 *
	 * @param src
	 *            the source expression to convert
	 * @param type
	 *            classfile descriptor, on the form <code>I</code> or
	 *            <code>Ljava/lang/String;</code>
	 * @return the source expression that will produce a value of the target type.
	 */
	private Expression convertToTargetType(Node source, Expression src, String type) {
		char firstChar = type.charAt(0);
		if (type.length() == 1) {
			switch (firstChar) {
			case 'B': {
				return newMessageSend("JS_Toboolean", new Expression[] { src });
			}
			case 'C': {
				return newMessageSend("JS_Tochar", new Expression[] { src });
			}
			case 'S': {
				return newMessageSend("JS_Toshort", new Expression[] { src });
			}
			case 'I': {
				return newMessageSend("JS_Toint", new Expression[] { src });
			}
			case 'J': {
				return newMessageSend("JS_Tolong", new Expression[] { src });
			}
			case 'F': {
				return newMessageSend("JS_Tofloat", new Expression[] { src });
			}
			case 'D': {
				return newMessageSend("JS_Todouble", new Expression[] { src });
			}
			default:
				throw new UnsupportedJavaScriptSyntax(source);
			}
		}
		else if (firstChar == 'L') {
			// cast
			String typeRef = type.substring(1, type.length() - 1);
			if (typeRef.equals("java/lang/String")) {
				return newMessageSend("JS_ToString", new Expression[] { src });
			} else {
				return new CastExpression(src, javaTypeReferenceFromClassfileType(source, typeRef));
			}
		}
		else {
			throw new AssertionError();
		}
	}

	/**
	 * @param call
	 * @return
	 */
	private char[] getGeneratedMethodName(InstanceMethodCall call) {
		return javaName("JS_instanceMethodCallFunction_" + call.getName() + "_" + call.getNbParameters());
	}

	private void addInstanceMethodCall(InstanceMethodCall call) {
		// linear search as we consider there will not have a big list of difference
		// method calls (10-100 methods)
		if (!instanceMethodsCalls.contains(call)) {
			instanceMethodsCalls.add(call);
		}
	}

	public InstanceMethodCall[] getInstanceMethodCalls() {
		return instanceMethodsCalls.toArray(new InstanceMethodCall[instanceMethodsCalls.size()]);
	}

	/**
	 * @param type
	 *            on the form <code>a/b/C</code> or <code>[[[La/b/C;</code>
	 * @return the {@link TypeReference} corresponding to the given type
	 */
	private TypeReference javaTypeReferenceFromClassfileType(Node source, String type) {
		int nbDimensions = 0;
		while (type.charAt(0) == '[') {
			++nbDimensions;
			type = type.substring(1, type.length());
		}
		if(nbDimensions > 0) {
			char firstChar = type.charAt(0);
			if(type.length() == 1) {
				// basetype
				throw new UnsupportedJavaScriptSyntax(source); // TODO
			}
			else if (firstChar == 'L') {
				// cast
				type = type.substring(1, type.length() - 2);
			}
			else {
				throw new AssertionError();
			}
		}

		TypeReference javaType = getTypeReferenceClassfile(type);
		javaType.dimensions = nbDimensions;

		return javaType;
	}

	/**
	 * Expand a call node. When returning true, a new expression has been pushed
	 *
	 * @param callNode
	 * @param functionName
	 * @param receiver
	 *            may be null
	 * @return true if the node has been expanded, false otherwise
	 */
	private boolean expandBuiltinNode(CallNode callNode, String functionName,
			jdk.nashorn.internal.ir.Expression receiver) {
		if (receiver == null) {
			// expand well known global object functions
			if (functionName.equals("nls")) {
				// expand nls(id)
				List<jdk.nashorn.internal.ir.Expression> args = callNode.getArgs();
				if (args.size() != 1) {
					throw new UnsupportedJavaScriptSyntax(callNode);
				}
				jdk.nashorn.internal.ir.Expression arg = args.get(0);
				if (!(arg instanceof LiteralNode)) {
					throw new UnsupportedJavaScriptSyntax(arg);
				}
				int nlsKeyId = ((LiteralNode) arg).getInt32();

				MessageSend getNLS = new MessageSend(javaName("getNLS"), javaFullyQualifiedName("ej.giml.GIML"), null,
						new Expression[0]);
				MessageSend getMessage = new MessageSend(javaName("getMessage"), javaFullyQualifiedName("ej.nls.NLS"),
						getNLS,
						new Expression[] { new IntegerLiteral(nlsKeyId) });
				pushExpression(getMessage);
				return true;
			}

			if (functionName.equals("setTimeout") || functionName.equals("setInterval")) {
				// set[Timeout|Interval](function, millis, args...)
				List<jdk.nashorn.internal.ir.Expression> args = callNode.getArgs();
				if (args.size() < 2) {
					throw new UnsupportedJavaScriptSyntax(callNode);
				}
				List<Expression> javaArgsVect = generateExpressions(args);

				Expression function = javaArgsVect.get(0);
				Expression millis = javaArgsVect.get(1);
				javaArgsVect = javaArgsVect.subList(2, javaArgsVect.size());
				TypeReference objectArray = getJavaLangObjectArray();
				Expression functionArgs = new ArrayAllocation(javaArgsVect.toArray(new Expression[javaArgsVect.size()]),
						objectArray);
				Expression javaCall = newMessageSend(functionName, new Expression[] { function, millis, functionArgs });
				pushExpression(javaCall);
				return true;
			}
			return false;
		} else {
			String fullyQualifiedNameExpression = getFullyQualifiedNameExpression(receiver);
			if (isSimpleName(fullyQualifiedNameExpression)) {
				String varName = fullyQualifiedNameExpression;
				if (!isLocalVariable(varName)) {
					GlobalObject go = javaScriptGenerator.getGlobalObjects().get(varName);
					if (go != null) {
						Expression expr = go.generateCallExpression(this, callNode, functionName);
						if (expr != null) {
							pushExpression(expr);
							return true;
						}
					}
				}
			}

			if (functionName.equals("call")) {
				receiver.accept(this);
				Expression receiverExpr = popExpression();
				Expression toFunction = newMessageSend("JS_ToFunction", new Expression[] { receiverExpr });

				List<jdk.nashorn.internal.ir.Expression> args = callNode.getArgs();
				ArrayList<Expression> javaArgsVect = new ArrayList<>();
				for (jdk.nashorn.internal.ir.Expression arg : args) {
					arg.accept(this);
					javaArgsVect.add(popExpression());
				}

				TypeReference functionType = getTypeReference("com.is2t.giml.js.GIMLJavaScriptCode.GIMLFunctionObject");
				TypeReference objectArray = getJavaLangObjectArray();
				ArrayAllocation exprArray = new ArrayAllocation(
						javaArgsVect.toArray(new Expression[javaArgsVect.size()]), objectArray);

				pushExpression(new MessageSend(javaName("call"),
						functionType,
						toFunction,
						new Expression[] { exprArray }));
				return true;
			}
			return false;
		}
	}

	private TypeReference getJavaLangObjectArray() {
		return new TypeReference(TYPE_OBJECT.fullyQualifiedName, 1);
	}

	private List<Expression> generateExpressions(List<jdk.nashorn.internal.ir.Expression> expressions) {
		ArrayList<Expression> javaArgsVect = new ArrayList<>();
		for (jdk.nashorn.internal.ir.Expression expr : expressions) {
			expr.accept(this);
			javaArgsVect.add(popExpression());
		}
		return javaArgsVect;
	}

	/**
	 * Tell whether the given variable is a local (in the context of the function
	 * being generated)
	 *
	 * @return true if it is a local, false otherwise
	 */
	private boolean isLocalVariable(String varName) {
		return false; // TODO manage variables context
	}

	/**
	 * Tells whether this name is a simple name (without dots - not a fully
	 * qualified name)
	 *
	 * @param name
	 *            the name to test
	 * @return true if this name is a simple name, false otherwise
	 */
	private boolean isSimpleName(String name) {
		return name.indexOf('.') == -1;
	}

	/**
	 * Generate receiver for {@link MessageSend}, {@link FieldAccess}...
	 *
	 * @param receiver
	 *            the receiver to be generated
	 */
	private void generateReceiver(jdk.nashorn.internal.ir.Expression receiver) {
		String fullyQualifiedNameExpression = getFullyQualifiedNameExpression(receiver);
		// approximation
		// - fullyQualified name with dots: direct java static call with
		// TypeReferenceExpression
		// - simpleName: as TypeReferenceExpression so far to pass tests
		if (fullyQualifiedNameExpression != null && fullyQualifiedNameExpression.indexOf('.') != -1) {
			pushExpression(new TypeReferenceExpression(
					new TypeReference(javaFullyQualifiedName(fullyQualifiedNameExpression), 0)));
		} else {
			// normal message send - generate the receiver expression
			receiver.accept(this);
		}
	}

	private Expression[] javaExpressionArray(ArrayList<Expression> javaArgsVect) {
		return javaArgsVect.toArray(new Expression[javaArgsVect.size()]);
	}

	/**
	 * Get the java char array fully qualified name from the given javascript fully
	 * qualified name
	 *
	 * @param className
	 *            may be empty, return an empty array
	 * @return java char array fully qualified name from the given javascript fully
	 *         qualified name
	 */
	public char[][] javaFullyQualifiedName(String className) {
		if (className.equals("")) {
			return new char[0][];
		} else {
			String[] strs = className.split("\\.");
			int length = strs.length;
			char[][] array = new char[length][];
			for (int i = length; --i >= 0;) {
				array[i] = javaName(strs[i]);
			}
			return array;
		}
	}

	/**
	 * Get the java char array from the given javascript name
	 *
	 * @param name
	 *            the javascript name
	 * @return the java name
	 */
	public char[] javaName(String name) {
		return name.toCharArray();
	}

	/**
	 * Gets the expected MicroEJ class fully qualified name that contains global
	 * object functions (as static methods)
	 *
	 * @return the class fully qualified name.
	 */
	public String getGlobalObjectMicroEJClassName() {
		return globalObjectMicroEJClassName;
	}

	/**
	 * Sets the expected MicroEJ class fully qualified name that contains global
	 * object functions (as static methods)
	 *
	 * @param globalObjectMicroEJClassName
	 *            the class fully qualified name
	 */
	public void setGlobalObjectMicroEJClassName(String globalObjectMicroEJClassName) {
		this.globalObjectMicroEJClassName = globalObjectMicroEJClassName;
	}

	@Override
	public final Node leaveBinaryNode(final BinaryNode binaryNode) {
		Expression right = popExpression();
		Expression left = popExpression();
		String jsfunction = null;
		boolean jsRelationalFunction = false;
		boolean jsRelationalLeftToRight = false;
		boolean jsRelationalNegateResult = false;
		int operator = 0;
		switch (binaryNode.tokenType()) {
		case ASSIGN:{
			// a = b
			return leaveAssignment(binaryNode, left, right);
		}
		case ASSIGN_ADD:
		case ASSIGN_BIT_AND:
		case ASSIGN_BIT_OR:
		case ASSIGN_BIT_XOR:
		case ASSIGN_DIV:
		case ASSIGN_MOD:
		case ASSIGN_MUL:
		case ASSIGN_SAR:
		case ASSIGN_SHL:
		case ASSIGN_SHR:
		case ASSIGN_SUB:{
			// a += b
			return leaveCompoundAssignment(binaryNode);
		}
		case ADD: {
			// a+b (Number & String)
			jsfunction = "JS_Addition";
			break;
		}
		case DIV: {
			operator = BinaryExpression.Divide;
			break;
		}
		case MOD: {
			operator = BinaryExpression.Remainder;
			break;
		}
		case MUL: {
			operator = BinaryExpression.Multiply;
			break;
		}
		case SAR: {
			operator = BinaryExpression.RightShift;
			break;
		}
		case SHL: {
			operator = BinaryExpression.LeftShift;
			break;
		}
		case SHR: {
			operator = BinaryExpression.UnsignedRightShift;
			break;
		}
		case SUB: {
			operator = BinaryExpression.Minus;
			break;
		}
		// case BIT_AND:{
		// operator = BinaryExpression.BitwiseAnd;
		// break;
		// }
		// case BIT_OR:{
		// operator = BinaryExpression.BitwiseOr;
		// break;
		// }
		// case BIT_XOR:{
		// operator = BinaryExpression.BitwiseXor;
		// break;
		// }
		case AND: {
			operator = BinaryExpression.And;
			return generateConditionalOrAndOperator(binaryNode, operator, left, right);
		}
		case OR: {
			operator = BinaryExpression.Or;
			return generateConditionalOrAndOperator(binaryNode, operator, left, right);
		}
		case EQ:{
			jsfunction = "JS_AbstractEqualityComparison";
			break;
		}
		case EQ_STRICT: {
			jsfunction = "JS_StrictEqualityComparison";
			break;
		}
		case GE: {
			jsRelationalFunction = true;
			jsRelationalLeftToRight = true;
			jsRelationalNegateResult = true;
			break;
		}
		case GT: {
			jsRelationalFunction = true;
			jsRelationalLeftToRight = false;
			jsRelationalNegateResult = false;
			break;
		}
		case LE: {
			jsRelationalFunction = true;
			jsRelationalLeftToRight = false;
			jsRelationalNegateResult = true;
			break;
		}
		case LT: {
			jsRelationalFunction = true;
			jsRelationalLeftToRight = true;
			jsRelationalNegateResult = false;
			break;
		}
		case NE: {
			jsfunction = "JS_NotAbstractEqualityComparison";
			break;
		}

		case COMMARIGHT:
		case COMMALEFT:
		case IN:
		case INSTANCEOF:
		case NE_STRICT:
		case BIND:{
			throw new UnsupportedJavaScriptSyntax(binaryNode);
		}
		default:
			throw new AssertionError();
		}

		Expression result;
		if(jsfunction != null) {
			result = newMessageSend(jsfunction, new Expression[] { left, right });
		}
		else if (jsRelationalFunction) {
			result = newMessageSend("JS_AbstractRelationalComparison",
					new Expression[] { left, right, newBooleanLiteral(jsRelationalLeftToRight),
							newBooleanLiteral(jsRelationalNegateResult) });
		}
		else {
			assert operator != -1;
			result = newFloatValue(
					new BinaryExpression(operator, toNumberAsfloatValue(left), toNumberAsfloatValue(right)));
		}
		pushExpression(result);
		return binaryNode;
	}

	private Node generateConditionalOrAndOperator(BinaryNode binaryNode, int operator, Expression left,
			Expression right) {
		left = convertToBooleanType(binaryNode, left);
		right = convertToBooleanType(binaryNode, right);
		pushExpression(new BinaryExpression(operator, left, right));
		return binaryNode;
	}

	/**
	 * Get an expression that converts the given expression to a Number
	 * (<code>JS_ToNumber</code>) as a float value expression
	 *
	 * @param the
	 *            expression Object expression to convert to float
	 * @return an expression which is a java float
	 */
	private Expression toNumberAsfloatValue(Expression expr) {
		return new MessageSend(javaName("floatValue"), TYPE_FLOAT,
				newMessageSend("JS_ToNumber", new Expression[] { expr }), new Expression[0]);
	}

	/**
	 * Get an expression that converts a java float to a Float object
	 *
	 * @param the
	 *            expression to convert to a Float
	 * @return expression which is a Float
	 */
	private Expression newFloatValue(Expression expr) {
		return new ObjectAllocation(TYPE_FLOAT, new Expression[] { expr });
	}

	/**
	 * Creates a new MessageSend to a function declared in the generated type
	 *
	 * @param methodName
	 *            the method name
	 * @param args
	 *            the arguments
	 * @return the {@link MessageSend} expression
	 */
	private Expression newMessageSend(String methodName, Expression[] args) {
		return new MessageSend(javaName(methodName), javaFullyQualifiedName(getGlobalObjectMicroEJClassName()), null,
				args);
	}

	@Override
	public boolean enterFunctionNode(FunctionNode functionNode) {
		return false; // do not generate this function right now, do not push any expression (see
		// enterVarNode)
	}

	@Override
	public boolean enterTernaryNode(TernaryNode ternaryNode) {
		ternaryNode.getTest().accept(this);
		Expression test = convertToBooleanType(ternaryNode, popExpression());
		ternaryNode.getTrueExpression().accept(this);
		Expression trueExpr = popExpression();
		ternaryNode.getFalseExpression().accept(this);
		Expression falseExpr = popExpression();
		pushExpression(new TernaryExpression(test, trueExpr, falseExpr, TYPE_OBJECT));
		return false; // node has been visited
	}

	@Override
	public boolean enterReturnNode(ReturnNode returnNode) {
		com.is2t.java.nodes.Statement statement;
		if(returnNode.hasExpression()) {
			returnNode.getExpression().accept(this);
			statement = new ReturnStatement(popExpression());
		}
		else {
			statement = new ReturnStatement();
		}
		lastGeneratedJavaStatement = statement;
		return false; // node has been visited
	}

	@Override
	public boolean enterIfNode(IfNode ifNode) {
		ifNode.getTest().accept(this);
		Expression test = convertToBooleanType(ifNode, popExpression());

		ifNode.getPass().accept(this);
		com.is2t.java.nodes.Statement thenStatement = lastGeneratedJavaStatement;
		com.is2t.java.nodes.Statement elseStatement;
		jdk.nashorn.internal.ir.Block fail = ifNode.getFail();
		if (fail != null) {
			fail.accept(this);
			elseStatement = lastGeneratedJavaStatement;
		}
		else {
			elseStatement = null;
		}

		lastGeneratedJavaStatement = new IfStatement(test, thenStatement, elseStatement);
		return false; // node has been visited
	}

	@Override
	public boolean enterForNode(ForNode forNode) {
		jdk.nashorn.internal.ir.Expression initForJS = forNode.getInit();
		Expression init;
		if (initForJS != null) {
			initForJS.accept(this);
			init = popExpression();
		} else {
			init = null;
		}
		JoinPredecessorExpression testForJS = forNode.getTest();
		Expression test;
		if (testForJS != null) {
			testForJS.accept(this);
			test = convertToBooleanType(forNode, popExpression());
		}
		else {
			test = null;
		}
		JoinPredecessorExpression modifyForJS = forNode.getModify();
		Expression modify;
		if (modifyForJS != null) {
			modifyForJS.accept(this);
			modify = popExpression();
		} else {
			modify = null;
		}

		forNode.getBody().accept(this);
		com.is2t.java.nodes.Statement body = lastGeneratedJavaStatement;
		ForStatement forStatement = new ForStatement(init == null ? new Expression[0] : new Expression[] { init }, test,
				modify == null ? new Expression[0] : new Expression[] { modify },
						body);
		lastGeneratedJavaStatement = forStatement;
		return false; // node has been visited
	}

	@Override
	public boolean enterBlock(Block block) {
		Statement currentStatement = this.currentStatement;
		ArrayList<com.is2t.java.nodes.Statement> generatedStatementVect = new ArrayList<>();
		for (Statement s : block.getStatements()) {
			generate(s);
			if (lastGeneratedJavaStatement != null) {
				generatedStatementVect.add(lastGeneratedJavaStatement);
			}
		}
		this.lastGeneratedJavaStatement = new com.is2t.java.nodes.Block(
				generatedStatementVect.toArray(new com.is2t.java.nodes.Statement[generatedStatementVect.size()]));
		this.currentStatement = currentStatement;
		return false; // node has been visited
	}

	@Override
	public boolean enterVarNode(VarNode varNode) {
		if (currentStatement != varNode) {
			throw new UnsupportedJavaScriptSyntax(varNode);
		}
		// declaration of a new method (new context outside the code currently being generated)
		// varNode.get

		jdk.nashorn.internal.ir.Expression init = varNode.getInit();
		if (init != null) {
			init.accept(this);
			if (expressionStack.size() == 0) {
				// a function has been generated
				return false;
			}
			else {
				Expression initExpr = popExpression();
				lastGeneratedJavaStatement = new Assignment(getVariableAccess(varNode.getAssignmentDest()), initExpr);
			}
		}
		else {
			// a var declaration without expression - let as is (variables have been all
			// declared at the beginning of the function)
		}

		return false; // already visited
	}

	private Expression getVariableAccess(IdentNode node) {
		String varName = node.getName();
		if (currentAnalyzedFunction != null && currentAnalyzedFunction.isLocal(varName)) {
			// local
			return new LocalAccess(new LocalDeclaration(TYPE_OBJECT, javaName(varName), null));
		} else {
			// global
			// register the global object (will be initialized to undefined)
			GlobalObject go;
			FunctionNode function = jsSourcePreGenerator.getFunction(varName);
			if (function != null) {
				go = new GlobalObjectFunction(varName, function);
			} else {
				go = new GlobalObjectUserDefined(varName);
			}
			javaScriptGenerator.addGlobalObject(go);
			return new FieldAccess(TYPE_THIS_CLASS.fullyQualifiedName, javaName(varName), null);
		}
	}

	private Node leaveCompoundAssignment(BinaryNode binaryNode) {
		throw new UnsupportedJavaScriptSyntax(binaryNode);
	}

	private Node leaveAssignment(BinaryNode binaryNode, Expression left, Expression right) {
		pushExpression(new Assignment(left, right));
		return binaryNode;
	}

	/**
	 * Pop the expression which is on top of stack
	 *
	 * @return the popped expression
	 */
	public com.is2t.java.nodes.Expression popExpression() {
		try {
			return expressionStack.remove(expressionStack.size() - 1);
		} catch (IndexOutOfBoundsException e) {
			throw new AssertionError(e);
		}
	}

	/**
	 * Push the expression on stack
	 *
	 * @param expr
	 *            the expression to be pushed
	 *
	 */
	public void pushExpression(com.is2t.java.nodes.Expression expr) {
		expressionStack.add(expr);
	}

	/**
	 * Get the fully qualified name of the given node if this expression is only
	 * composed of names.
	 *
	 * @return null if the expression is not a fully qualified name
	 */
	private String getFullyQualifiedNameExpression(jdk.nashorn.internal.ir.Expression node) {
		if (node instanceof IdentNode) {
			return ((IdentNode) node).getName();
		} else if (node instanceof AccessNode) {
			AccessNode accessNode = (AccessNode) node;
			String baseName = getFullyQualifiedNameExpression(accessNode.getBase());
			if(baseName == null) {
				return null;
			}
			else {
				return baseName + '.'+ accessNode.getProperty();
			}
		}
		else {
			// this is a complex expression, not only composed of names
			return null;
		}
	}

	public com.is2t.java.nodes.Statement generateEmptyBlockStatement() {
		return new com.is2t.java.nodes.Block(new com.is2t.java.nodes.Statement[0]);
	}

	/**
	 * @return the expression to access to the "undefined object" field
	 */
	public Expression getUndefinedObject() {
		// shall be in sync with the field defined in the template
		return new FieldAccess(TYPE_THIS_CLASS.fullyQualifiedName, javaName("JS_UNDEFINED_OBJECT"), null);
	}

	/**
	 * @param fullyQualifiedName
	 *            on the form a.b.C
	 * @return
	 */
	public TypeReference getTypeReference(String fullyQualifiedName) {
		return new TypeReference(javaFullyQualifiedName(fullyQualifiedName), 0);
	}

	public void generateImplicitReturn(List<com.is2t.java.nodes.Statement> allJavaStatements) {
		int lastJavaStatementIndex = allJavaStatements.size() - 1;
		if (lastJavaStatementIndex == -1) {
			allJavaStatements.add(getReturnUndefinedObject());
		} else {
			com.is2t.java.nodes.Statement lastJavaStatement = allJavaStatements.get(lastJavaStatementIndex);
			if (lastJavaStatement instanceof Expression && !isVoidFunction(lastJavaStatement)) {
				// automatic return the last expression
				allJavaStatements.remove(lastJavaStatementIndex);
				allJavaStatements.add(new ReturnStatement(((Expression) lastJavaStatement)));
			}
			else if (!(lastJavaStatement instanceof ReturnStatement)) {
				allJavaStatements.add(getReturnUndefinedObject());
			}
		}
	}

	private ReturnStatement getReturnUndefinedObject() {
		return new ReturnStatement(getUndefinedObject());
	}

	private boolean isVoidFunction(com.is2t.java.nodes.Statement lastJavaStatement) {
		// TODO to be connected to classfile analysis
		if(lastJavaStatement instanceof MessageSend) {
			MessageSend messageSend = (MessageSend) lastJavaStatement;
			if (messageSend.toString().startsWith("System.out.print")) {
				return true;
			}
		}
		return false;
	}

	public MethodDeclaration generateInstanceMethodCall(ClasspathModel classpathModel, InstanceMethodCall call) {
		ArrayList<LocalDeclaration> paramsVect = new ArrayList<>();
		char[] javaNameThis = javaName("this_");
		LocalDeclaration localDeclarationThis = new LocalDeclaration(TYPE_OBJECT, javaNameThis, null);
		paramsVect.add(localDeclarationThis);
		int nbParameters = call.getNbParameters();
		for (int i = 0; i < nbParameters; ++i) {
			paramsVect.add(new LocalDeclaration(TYPE_OBJECT, javaName("arg"+i), null));
		}
		LocalDeclaration[] parameters = paramsVect.toArray(new LocalDeclaration[paramsVect.size()]);

		List<com.is2t.java.nodes.Statement> generatedStatementsVect = new ArrayList<>();

		String callMethodName = call.getName();
		InstanceMethod[] methods = classpathModel.getAllInstanceMethods(callMethodName, nbParameters);
		for (InstanceMethod method : methods) {
			// Generate something like:
			// if(this_ instanceof <TYPE>){
			// return ((<TYPE>)this_).<NAME>(map(arg0), ... map(argN);
			// ...
			// where map(arg) is JS_To<basetype>(arg) | JS_ToString(arg) | ((<ARG_TYPE)arg)
			TypeReference type = getTypeReferenceClassfile(method.getType());
			LocalAccess thisLocalAccess = new LocalAccess(localDeclarationThis);
			Expression instanceofExpr = new BinaryExpression(BinaryExpression.Instanceof,
					thisLocalAccess, new TypeReferenceExpression(type));

			Expression[] argsExpressions = new Expression[nbParameters];
			String[] parameterDescriptors = method.getParameterDescriptors();
			for (int i = 0; i < nbParameters; ++i) {
				String argType = parameterDescriptors[i];
				argsExpressions[i] = convertToTargetType(null, new LocalAccess(parameters[i + 1]), argType);
			}

			List<com.is2t.java.nodes.Statement> statementsVect = new ArrayList<>();

			Expression callExpr = new MessageSend(javaName(callMethodName), type,
					new CastExpression(thisLocalAccess, type), argsExpressions);

			if(method.getReturnDescriptors().equals("V")){
				statementsVect.add(callExpr);
				statementsVect.add(getReturnUndefinedObject());
			}
			else {
				statementsVect.add(new ReturnStatement(callExpr));
			}

			com.is2t.java.nodes.Block ifBlock = new com.is2t.java.nodes.Block(statementsVect.toArray(new com.is2t.java.nodes.Statement[statementsVect.size()]));
			IfStatement ifStatement = new IfStatement(instanceofExpr, ifBlock);
			generatedStatementsVect.add(ifStatement);
		}
		generatedStatementsVect.add(new ThrowStatement(getNewJSTypeException()));

		com.is2t.java.nodes.Statement[] generatedStatements = generatedStatementsVect
				.toArray(new com.is2t.java.nodes.Statement[generatedStatementsVect.size()]);

		MethodDeclaration md = new MethodDeclaration(new ClassDeclaration(new char[0][]),
				new Modifiers(Modifiers.AccPublic | Modifiers.AccStatic),
				getGeneratedMethodName(call), parameters, TYPE_OBJECT,
				generatedStatements);

		return md;
	}

	private Expression getNewJSTypeException() {
		return new ObjectAllocation(getTypeReference("JSTypeErrorException"), new Expression[0]);
	}

	/**
	 * @param type
	 *            classfile type reference on the form a/b/C$D
	 * @return the {@link TypeReference}
	 */
	private TypeReference getTypeReferenceClassfile(String type) {
		return getTypeReference(type.replace('$', '.').replace('/', '.'));
	}
}