/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Stack;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.xml.sax.SAXException;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.model.Template;
import com.is2t.giml.xml.parser.exception.XMLParserCircularImportException;
import com.is2t.giml.xml.parser.exception.XMLParserDuplicatedDocumentException;
import com.is2t.giml.xml.parser.exception.XMLParserException;
import com.is2t.giml.xml.parser.exception.XMLParserMalformedTemplateException;
import com.is2t.giml.xml.parser.exception.XMLParserMisplacedElementException;
import com.is2t.giml.xml.parser.exception.XMLParserUnknownAttributeException;
import com.is2t.giml.xml.parser.exception.XMLParserUnknownDocumentException;
import com.is2t.giml.xml.parser.utils.FileUtils;

/**
 * The class {@code XMLParser} represent a generic XML file parser that can be configured with an instance of
 * {@link XMLParserOptions}. Parser has support for <code>import</code> directives.
 *
 * @see XMLParserOptions
 * @see XMLParserConstants#DEFAULT_IMPORT_DOCUMENT_TAGNAME
 */
public class XMLParser {

	private final InputStream _rootInput; // NOSONAR
	private final List<File> _workingDirectories; // NOSONAR
	private XMLParserOptions _options; // NOSONAR

	private String _rootDocumentId; // NOSONAR
	private final Map<String, Document> _allDocuments; // NOSONAR
	private final Stack<Document> _processingDocuments; // NOSONAR
	private final Map<String, Document> _resolvedDocuments; // NOSONAR

	private final List<String> _knownTemplates; // NOSONAR

	/**
	 * Creates a parser with default options that will parse the given input stream. If parser must process import tags,
	 * parser will look for documents to import in the given working directory.
	 *
	 * @param inputStream
	 *            the root document to parse.
	 * @param workingDirectories
	 *            the working directory to look for imported modules.
	 * @throws NullPointerException
	 *             the given document is <code>null</code>.
	 * @throws IllegalArgumentException
	 *             if the given working directory is <code>null</code>, does not exist or does not denote a directory.
	 *
	 * @see File#isFile()
	 * @see File#isDirectory()
	 */
	public XMLParser(InputStream inputStream, List<File> workingDirectories)
			throws NullPointerException, IllegalArgumentException {
		if(workingDirectories == null || workingDirectories.size() == 0){
			throw new IllegalArgumentException("No directory to look for imported modules");
		}
		FileUtils.checkValidDirectory(workingDirectories);

		this._rootInput = inputStream;
		this._workingDirectories = workingDirectories;
		this._options = new XMLParserOptions();

		this._allDocuments = new HashMap<String, Document>();
		this._processingDocuments = new Stack<Document>();
		this._resolvedDocuments = new HashMap<String, Document>();

		this._knownTemplates = new ArrayList<String>();
	}

	/**
	 * Creates a parser with default options that will parse the given file. If parser must process import tags, parser
	 * will look for documents to import in the given file parent folder and sub-folders.
	 *
	 * @param rootFile
	 *            the root document to parse.
	 * @throws NullPointerException
	 *             the the given document is <code>null</code>.
	 * @throws FileNotFoundException
	 *             if the given file does not exist or does not denote a file.
	 *
	 * @see File#isFile()
	 */
	public XMLParser(File rootFile) throws NullPointerException, FileNotFoundException {
		this._rootInput = new FileInputStream(rootFile);
		File workingDirectory = rootFile.getParentFile();
		FileUtils.checkValidDirectory(workingDirectory);
		List<File> workingDirectories = new ArrayList<File>(1);
		workingDirectories.add(workingDirectory);
		this._workingDirectories = workingDirectories;
		this._options = new XMLParserOptions();

		this._allDocuments = new HashMap<String, Document>();
		this._processingDocuments = new Stack<Document>();
		this._resolvedDocuments = new HashMap<String, Document>();

		this._knownTemplates = new ArrayList<String>();
	}

	/**
	 * Creates a parser with default options that will parse the given file. If parser must process import tags, parser
	 * will look for documents to import in the given working directory.
	 *
	 * @param rootFile
	 *            the root document to parse.
	 * @param workingDirectories
	 *            the working directories to look for imported modules.
	 * @throws NullPointerException
	 *             the given document is <code>null</code>.
	 * @throws FileNotFoundException
	 *             if the given file does not exist or does not denote a file.
	 * @throws IllegalArgumentException
	 *             if the given working directory is <code>null</code>, does not exist or does not denote a directory.
	 *
	 * @see File#isFile()
	 * @see File#isDirectory()
	 */
	public XMLParser(File rootFile, List<File> workingDirectories)
			throws NullPointerException, FileNotFoundException, IllegalArgumentException {
		this(new FileInputStream(rootFile), workingDirectories);
	}

	/**
	 * Gets parser options.
	 *
	 * @return parser options.
	 */
	public XMLParserOptions getOptions() {
		return this._options;
	}

	/**
	 * Sets options for this parser.
	 *
	 * @param options
	 *            new options.
	 * @throws IllegalArgumentException
	 *             if given options are <code>null</code>.
	 */
	public void setOptions(XMLParserOptions options) throws IllegalArgumentException {
		if (options == null) {
			throw new IllegalArgumentException("Parser options cannot be null");
		}
		this._options = options;
	}

	/**
	 * Parses given documents to get an tree-like representation of XML file content.
	 *
	 * @return a tree-like representation of the root file.
	 * @throws XMLParserException
	 *             if something wrong happened.
	 */
	public Document parse() throws XMLParserException {
		if (!this._options.handleImports()) {
			// Parse must not handle imports. So the result of parsing all file is just the result of parsing the root
			// file.
			return this.parseXMLFile(this._rootInput);
		}

		// To parse all documents and resolve imports:
		// First, parse all XML files to extract data and create a model approximation
		// this approximation does not handle import tags, that is to say, import tags are kept but not interpreted
		List<File> workingFiles = FileUtils.listFiles(this._workingDirectories, this._options.getXMLFileExtension());
		this.resolveDocumentsWithoutImports(this._rootInput, workingFiles);

		// Then, we now have access to all documents, and their IDs. It is now possible to resolve import directives
		// inside documents
		// for (Document document : this._allDocuments.values()) {
		// this.checkForMisplacedElement(document, this._options.getImportTagName());
		// this.resolveDocumentsWithImports(document);
		// }
		Document rootDocument = this._allDocuments.get(this._rootDocumentId);
		this.checkForMisplacedElement(rootDocument, this._options.getImportTagName());
		this.resolveDocumentsWithImports(rootDocument);

		// Look for templates if parser must handle templates
		if (this._options.handleTemplates()) {
			this.extractTemplatesFromDocument(this._rootDocumentId);
		}

		// Result is the resolved root document. Do not compact it because if user needs nested document information, he
		// cannot have it once it is root is compacted
		return this._resolvedDocuments.get(this._rootDocumentId);
	}

	private Document parseXMLFile(InputStream xmlInput) throws XMLParserException {
		try {
			DocumentBuilder docBuilder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
			org.w3c.dom.Document xmlDocument = docBuilder.parse(xmlInput);
			xmlDocument.getDocumentElement().normalize();
			return XMLParserModelFactory.createDocument(xmlDocument.getDocumentElement(),
					this._options.getDocumentIdAttributeName());
		} catch (ParserConfigurationException | IOException | SAXException e) {
			throw new XMLParserException(e.getMessage(), e);
		}
	}

	private void resolveDocumentsWithoutImports(InputStream xmlRoot, List<File> xmlFiles) throws XMLParserException {
		for (File xmlFile : xmlFiles) {
			try {
				Document document = this.parseXMLFile(new FileInputStream(xmlFile));
				// Before adding document to the document pool, check if a document with the same ID already exist.
				if (this._allDocuments.containsKey(document.getId())) {
					throw new XMLParserDuplicatedDocumentException(document.getId());
				}
				this._allDocuments.put(document.getId(), document);
			} catch (FileNotFoundException e) {
				// Should not occur as long as we browse existing files in the working directory.
				throw new XMLParserException("File not found", e);
			}
		}
		Document rootDocument = this.parseXMLFile(xmlRoot);
		this._rootDocumentId = rootDocument.getId();
		this._allDocuments.put(this._rootDocumentId, rootDocument);
	}

	private Document resolveDocumentsWithImports(Document originalDocument) throws XMLParserException {
		// First action is to push document to resolve into list of processing documents. It is important to do that to
		// detect circular imports when resolving import directive: no infinite loop > happy user
		this._processingDocuments.add(originalDocument);

		// To get a document with resolved imports, we recreated a new document, that is a copy of the original
		// document, and we loop over original document node, and:
		// - if node is an import node, we resolve the import using original documents pool (from first parse)
		// - once resolved, we evict import node since it is not useful anymore
		// After that, the new document is created with the new list of child nodes and documents
		List<Node> childNodes = new ArrayList<Node>();
		List<Document> childDocuments = new ArrayList<Document>();

		String importTagName = this._options.getImportTagName();
		String importTagAttributeName = this._options.getImportTagAttributeName();

		for (Node originalChild : originalDocument.getChildren()) {
			if (originalChild.getName().equals(importTagName)) {
				// The original child node is an import directive, check if the import tag is well formed
				Attribute importedDocumentIdAttribute = originalChild.getAttribute(importTagAttributeName);
				if (importedDocumentIdAttribute == null) {
					throw new XMLParserUnknownAttributeException(importTagAttributeName, importTagName,
							originalDocument.getId());
				}
				// Import is well formed, try to resolve nested documents
				this.resolveImport(childDocuments, importTagName, importedDocumentIdAttribute);
			} else {
				// This child node is not an import directive, keep this node
				childNodes.add(originalChild);
			}
		}
		this._processingDocuments.remove(originalDocument);
		Document resolvedDocument = XMLParserModelFactory.createDocument(originalDocument.getId(),
				originalDocument.getName(), originalDocument.getAttributes(), childNodes, childDocuments);
		this._resolvedDocuments.put(resolvedDocument.getId(), resolvedDocument);
		return resolvedDocument;
	}

	private void resolveImport(List<Document> childDocuments, String importTagName,
			Attribute importedDocumentIdAttribute) throws XMLParserException {
		// First, check if the imported document exists, i.e. check if we know it in our original document pool.
		String importedDocumentId = importedDocumentIdAttribute.getValue();
		Document docToImport = this._allDocuments.get(importedDocumentId);
		if (docToImport == null) {
			throw new XMLParserUnknownDocumentException(importedDocumentId);
		}

		// Here, document exists for sure, 4 possible cases:
		// - the document to include is not a leaf, and is already in our processing document pool: this means there is
		// a circular import
		// - the document to include is not a leaf, but is already resolved, i.e. it is already in our resolved document
		// pool
		// - the document to include is a leaf document, i.e. it does not include other documents
		// - the document to include is not a leaf, is not processing and has not already been processed: process it and
		// at it to the document children
		if (this._processingDocuments.contains(docToImport)) {
			// document not processed but already in processing queue: circular reference
			throw XMLParserCircularImportException.fromCircularDependencyDetails(this._processingDocuments);
		} else if (this._resolvedDocuments.containsKey(docToImport.getId())) {
			// document already resolved, add it but do not go recursively
			childDocuments.add(docToImport);
			return;
		} else if (!docToImport.hasChild(importTagName)) {
			// Not processed but has no import > consider it is processed
			this._resolvedDocuments.put(importedDocumentId, docToImport);
			childDocuments.add(docToImport);
		} else {
			// recursive processing
			childDocuments.add(this.resolveDocumentsWithImports(docToImport));
		}
	}

	private Document extractTemplatesFromDocument(String documentId)
			throws XMLParserMalformedTemplateException, XMLParserMisplacedElementException {
		// Always get the latest version of the document from the resolved document pool
		Document document = this._resolvedDocuments.get(documentId);
		this.checkForMisplacedElement(document, this._options.getTemplateTagName());

		List<Document> referencedDocuments = new ArrayList<Document>();
		for (Document referencedDocument : document.getReferencedDocuments()) {
			referencedDocuments.add(this.extractTemplatesFromDocument(referencedDocument.getId()));
		}

		List<Template> templates = new ArrayList<Template>();
		List<Node> children = new ArrayList<Node>();
		String templateTagName = this._options.getTemplateTagName();
		String templateTagAttributeName = this._options.getTemplateTagAttributeName();
		String templateParameterPrefix = this._options.getTemplateParameterPrefix();
		String templateParameterSuffix = this._options.getTemplateParameterSuffix();
		String templateContentTagName = this._options.getTemplateContentTagName();
		String templateContentTagAttributeName = this._options.getTemplateContentTagAttributeName();
		String templateContentTagInstantiationAttributeName = this._options
				.getTemplateContentTagInstantiationAttributeName();

		try {
			// Templates are necessarily children of root element
			for (Node child : document.getChildren()) {
				if (templateTagName.equals(child.getName())) {
					Template template = new Template(child, templateTagName, templateTagAttributeName,
							templateParameterPrefix, templateParameterSuffix, templateContentTagName,
							templateContentTagAttributeName, templateContentTagInstantiationAttributeName);
					String templateName = template.getName();
					if (this._knownTemplates.contains(templateName)) {
						throw new XMLParserMalformedTemplateException(
								String.format("Template '%s' already exists.", templateName));
					}
					templates.add(template);
					this._knownTemplates.add(templateName);
				} else {
					children.add(child);
				}
			}

			Document resolvedDocument = XMLParserModelFactory.createDocument(document.getId(), document.getName(),
					document.getAttributes(), children, referencedDocuments, templates);
			this._resolvedDocuments.put(resolvedDocument.getId(), resolvedDocument);
			return resolvedDocument;
		} catch (NullPointerException | IllegalArgumentException e) {
			throw new XMLParserMalformedTemplateException("Cannot resolve template", e);
		}
	}

	private void checkForMisplacedElement(Node document, String elementTagName)
			throws XMLParserMisplacedElementException {
		// This check indirection allows to put import tag under document root but not in children
		for (Node child : document.getChildren()) {
			if (child.hasChild(elementTagName)) {
				throw new XMLParserMisplacedElementException(elementTagName, child.getName());
			}
			this.checkForMisplacedElement(child, elementTagName);
		}
	}

}
