/*
 * Java
 *
 * Copyright 2016 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.giml.xml.parser.test.xmlparser;

import java.io.FileNotFoundException;
import java.util.Set;

import org.junit.Assert;
import org.junit.Test;

import com.is2t.giml.xml.model.Attribute;
import com.is2t.giml.xml.model.Document;
import com.is2t.giml.xml.model.Node;
import com.is2t.giml.xml.parser.XMLParser;
import com.is2t.giml.xml.parser.XMLParserOptions;
import com.is2t.giml.xml.parser.exception.XMLParserException;

/**
 * Test cases of {@link XMLParser} on valid documents and resolved imports.
 */
public class XMLParserValidDocumentsResolvedImportsTest extends AbstractXMLParserValidTest {

	private static final String RESOURCE_FOLDER_PATH = "/com/is2t/giml/xml/parser/test/xmlparser/imports/valid/";

	private static final String PARSER_OPTION_XML_EXTENSION = ".xml";
	private static final String PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME = "name";
	private static final String PARSER_OPTION_IMPORT_TAGNAME = "import";
	private static final String PARSER_OPTION_IMPORT_ATTRIBUTE_NAME = "name";
	private static final boolean PARSER_OPTION_HANDLE_IMPORTS = true;

	@Override
	protected XMLParserOptions getXMLParserOptions() {
		XMLParserOptions options = new XMLParserOptions();
		options.setXMLFileExtension(PARSER_OPTION_XML_EXTENSION);
		options.setDocumentIdAttributeName(PARSER_OPTION_DOCUMENT_ID_ATTRIBUTE_NAME);
		options.setImportTagName(PARSER_OPTION_IMPORT_TAGNAME);
		options.setImportTagAttributeName(PARSER_OPTION_IMPORT_ATTRIBUTE_NAME);
		options.setHandleImports(PARSER_OPTION_HANDLE_IMPORTS);
		return options;
	}

	@Override
	protected String getResourceFolder() {
		return RESOURCE_FOLDER_PATH;
	}

	@Test
	public void testSimpleDocument() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("simple/simple.xml")).parse();

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "simpleId");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 2);
		Assert.assertEquals(document.getReferencedDocumentCount(), 0);

		// Check first child
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check second child
		Node subnode2 = createNode("simplesubnode2", NO_ATTRIBUTE);
		Node node2 = createNode("simplenode2", NO_ATTRIBUTE, subnode2);
		Assert.assertTrue(document.getChildren().contains(node2));
	}

	@Test
	public void testSimpleDocumentsWithImport() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("simpleimport/doc1.xml")).parse();

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "doc1");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 2);
		Assert.assertEquals(document.getReferencedDocumentCount(), 1);
		Assert.assertFalse(document.hasChild(PARSER_OPTION_IMPORT_TAGNAME)); // import are resolved

		// Check first child
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check second child
		Node subnode2 = createNode("simplesubnode2", NO_ATTRIBUTE);
		Node node2 = createNode("simplenode2", NO_ATTRIBUTE, subnode2);
		Assert.assertTrue(document.getChildren().contains(node2));

		// Resolve imported document as a simple document and check root document contains the newly parsed doc
		Document importedDocument = createXMLParser(getFileFromJarResource("simpleimport/doc2.xml")).parse();
		Assert.assertTrue(document.getReferencedDocuments().contains(importedDocument));
	}

	@Test
	public void testDocumentsWithChainImport() throws XMLParserException, FileNotFoundException {
		Document document = createXMLParser(getFileFromJarResource("chainimports/doc1.xml")).parse();
		document = document.compact();

		System.out.println(document);

		// Check root
		Assert.assertNotNull(document);
		Assert.assertEquals(document.getName(), "root");
		Assert.assertEquals(document.getId(), "doc1");
		Assert.assertEquals(document.getAttributeCount(), 1); // the ID is an attribute
		Assert.assertEquals(document.getChildCount(), 4); // one per file
		Assert.assertEquals(document.getReferencedDocumentCount(), 0); // compacted document
		Assert.assertFalse(document.hasChild(PARSER_OPTION_IMPORT_TAGNAME)); // import are resolved

		// Check doc1 node
		Set<Attribute> node1AttrSet = createAttributeSet("attribute1", "value1", "attribute2", "value2");
		Node node1 = createNode("simplenode1", node1AttrSet);
		Assert.assertTrue(document.getChildren().contains(node1));

		// Check doc2 node
		Set<Attribute> node2AttrSet = createAttributeSet("attribute2", "value2");
		Node node2 = createNode("simplenode2", node2AttrSet);
		Assert.assertTrue(document.getChildren().contains(node2));

		// Check doc3 node
		Set<Attribute> node3AttrSet = createAttributeSet("attribute3", "value3");
		Node node3 = createNode("simplenode3", node3AttrSet);
		Assert.assertTrue(document.getChildren().contains(node3));

		// Check doc4 node
		Set<Attribute> node4AttrSet = createAttributeSet("attribute4", "value4");
		Node node4 = createNode("simplenode4", node4AttrSet);
		Assert.assertTrue(document.getChildren().contains(node4));
	}

}
