/*
 * Java
 *
 * Copyright 2009-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.is2t.microbsp.microui.geg;

import ej.microui.MicroUIPump;
import ej.microui.event.EventHandler;
import ej.microui.event.generator.Pointer;

/**
 * Instance created by SystemMicroUIGenerator
 */
public class SystemPointer extends Pointer {

	/**
	 * Driver MicroEJ specific:
	 * <p>
	 * event0: PPPP PPPP EEEE EEEE xxxx xxxx xxxx xxxx<br>
	 * event1: ---t cccc bbbb bbbb yyyy yyyy yyyy yyyy<br>
	 * <p>
	 * where
	 * <ul>
	 * <li>P is the pump native event (0x7)</li>
	 * <li>E it the EventGenerator id</li>
	 * <li>t is the type absolute or relative</li>
	 * <li>x is x</li>
	 * <li>c is the pointer command</li>
	 * <li>b is the button id</li>
	 * <li>y is y</li>
	 * </ul>
	 */
	// masks & shifts
	private static final int DRIVER_TYPE_SHIFT = 28;
	private static final int DRIVER_TYPE_MASK = 0x1 << DRIVER_TYPE_SHIFT;
	private static final int DRIVER_CMD_SHIFT = 24;
	private static final int DRIVER_CMD_MASK = 0xf << DRIVER_CMD_SHIFT;
	private static final int DRIVER_BUTTON_SHIFT = 16;
	private static final int DRIVER_BUTTON_MASK = 0xff << DRIVER_BUTTON_SHIFT;
	private static final int DRIVER_DATA_SHIFT = 0;
	private static final int DRIVER_DATA_MASK = 0xffff << DRIVER_DATA_SHIFT;

	// types values
	private static final int DRIVER_TYPE_DELTA = 0x1; // 0 means ABS

	// cmd values
	private static final int DRIVER_CMD_MOVE = 0x0;
	private static final int DRIVER_CMD_PRESS = 0x1;
	private static final int DRIVER_CMD_RELEASE = 0x2;
	private static final int DRIVER_CMD_LONG = 0x3;
	private static final int DRIVER_CMD_REPEAT = 0x4;

	/**
	 * There is only one pointer event converter per native pointer. if the given id
	 * is invalid or has been already set, throw runtime
	 *
	 * @param nbButtons
	 *            available number of extended buttons
	 * @param width
	 *            area width
	 * @param height
	 *            area height
	 */
	public SystemPointer(int nbButtons, int width, int height) {
		super(nbButtons, width, height);
	}

	@Override
	public void convert(MicroUIPump pump, int inputEvent, EventHandler eventHandler) throws InterruptedException {

		// manage x coordinate & absolute or relative
		int x = (inputEvent & DRIVER_DATA_MASK) >>> DRIVER_DATA_SHIFT;

		// read next event
		inputEvent = pump.read();

		int y = (inputEvent & DRIVER_DATA_MASK) >>> DRIVER_DATA_SHIFT;

		if (((inputEvent & DRIVER_TYPE_MASK) >>> DRIVER_TYPE_SHIFT) == DRIVER_TYPE_DELTA) {
			x = getAbsoluteX() + (short) x; // x is on 16 bits for sure
			y = getAbsoluteY() + (short) y; // y is on 16 bits for sure
		}

		int cmd = (inputEvent & DRIVER_CMD_MASK) >>> DRIVER_CMD_SHIFT;

		switch (cmd) {
		case DRIVER_CMD_MOVE:
			move(x, y);
			return; // nothing else to do
		case DRIVER_CMD_PRESS:
			cmd = PRESSED;
			break;
		case DRIVER_CMD_RELEASE:
			cmd = RELEASED;
			break;
		case DRIVER_CMD_REPEAT:
			cmd = REPEATED;
			break;
		case DRIVER_CMD_LONG:
			cmd = LONG;
			break;
		default:
			return; // error ?
		}

		// button event for sure
		// update pointer position
		if (cmd != RELEASED) {
			reset(x, y);
		}
		int id = (inputEvent & DRIVER_BUTTON_MASK) >>> DRIVER_BUTTON_SHIFT;
		send(cmd, id, eventHandler);
	}
}
