/*
 * Java
 *
 * Copyright 2016-2017 IS2T. All rights reserved.
 * IS2T PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package com.is2t.nls.generator;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;

import com.is2t.nls.StringEncoder;
import com.is2t.nls.binary.WatchableByteArrayOutputStream;

/**
 *
 */
public class NLSGenerator {

	/**
	 * The StringTemplate file name for used for source code generation
	 */
	private static final String NLS_STRING_TEMPLATE_FILE_NAME = "com/is2t/nls/generator/nls.stg";

	private static final String HEADER = ".header";
	private static final String BINARY_FILE = ".binary.file";
	private static final String SYSTEM_PROPERTIES = ".system.properties";
	private static final String RESOURCES_LIST = ".resources.list";
	private static final String IMMUTABLES_LIST = ".immutables.list";
	private static final String BIN = ".bin";
	private static final String XML = ".xml";
	private static final String JAVA = ".java";
	private static final String PO = ".po";

	private static final STGroup NlsSTGroup = new STGroupFile(NLS_STRING_TEMPLATE_FILE_NAME);

	private final StringEncoder stringEncoder;
	private final String header;
	private final String packageName;
	private final String packageDir;
	private final NLSLocale locales[];
	private final Map<String, Integer> messageKeyIds;
	private final File nlsGeneratedDir;
	private final File nlsGeneratedPropertiesDir;
	private WatchableByteArrayOutputStream out;
	private Map<String, Integer> positions;
	private boolean interfaceNameToUpperCase;

	public NLSGenerator(Set<NLSLocale> locales, Map<String, Integer> messageKeyIds, File outputDir,
			StringEncoder stringEncoder, String header, String packageName) throws NLSGeneratorException {
		if (locales == null || messageKeyIds == null || outputDir == null || stringEncoder == null) {
			throw new IllegalArgumentException("null argument");
		}
		this.stringEncoder = stringEncoder;
		this.header = header;
		this.packageName = getPackageName(packageName);
		this.packageDir = getPackageDir(packageName);
		this.locales = sortNLSLocales(locales);
		this.messageKeyIds = sortMessageKeyByIds(messageKeyIds);
		validateMessageIds();
		this.nlsGeneratedDir = new File(outputDir, this.packageDir);
		this.nlsGeneratedPropertiesDir = new File(outputDir, "properties"); //$NON-NLS-1$
		createOutputGeneratedDir(this.nlsGeneratedDir);
		createOutputGeneratedDir(this.nlsGeneratedPropertiesDir);
		this.interfaceNameToUpperCase = true; // backward compatible
	}

	/**
	 * Configures how to generate the name of the interface. By default this name is upper-cased.
	 *
	 * @param toUpperCase
	 *            false to disable the default behavior
	 */
	public void configureInterfaceNameCase(boolean toUpperCase) {
		this.interfaceNameToUpperCase = toUpperCase;
	}

	private String getPackageName(String packageName) {
		packageName = packageName.replaceAll("/", "\\."); //$NON-NLS-1$//$NON-NLS-2$
		if (packageName.endsWith(".")) { //$NON-NLS-1$
			packageName.substring(0, packageName.length() - 1);
		}
		if (packageName.startsWith(".")) { //$NON-NLS-1$
			packageName.substring(1);
		}
		return packageName;
	}

	private String getPackageDir(String packageName) {
		packageName = packageName.replaceAll("\\.", "/"); //$NON-NLS-1$ //$NON-NLS-2$
		if (!packageName.endsWith("/")) { //$NON-NLS-1$
			packageName = packageName + "/"; //$NON-NLS-1$
		}
		if (!packageName.startsWith("/")) { //$NON-NLS-1$
			packageName = "/" + packageName; //$NON-NLS-1$
		}
		return packageName;
	}

	private NLSLocale[] sortNLSLocales(Set<NLSLocale> locales) {
		NLSLocale[] localesArray = new NLSLocale[locales.size()];
		int i = 0;
		for (NLSLocale locale : locales) {
			localesArray[i] = locale;
			++i;
		}

		Arrays.sort(localesArray, new Comparator<NLSLocale>() {
			@Override
			public int compare(NLSLocale o1, NLSLocale o2) {
				return o1.getLocaleName().compareTo(o2.getLocaleName());
			}
		});

		return localesArray;
	}

	private void validateMessageIds() throws NLSGeneratorException {
		for (Entry<String, Integer> entry : this.messageKeyIds.entrySet()) {
			if (entry.getValue().intValue() <= 0) {
				throw new NLSGeneratorException("NLS message ID must be greater or equal than 1");
			}
		}
	}

	private int getMaxId() {
		int maxId = 0;
		for (Entry<String, Integer> entry : this.messageKeyIds.entrySet()) {
			int id = entry.getValue().intValue();
			if (entry.getValue().intValue() > maxId) {
				maxId = id;
			}
		}
		return maxId;
	}

	private String getMessageKeyById(Integer msgId) {
		for (Entry<String, Integer> entry : this.messageKeyIds.entrySet()) {
			if (entry.getValue().equals(msgId)) {
				return entry.getKey();
			}
		}
		return null;
	}

	private Map<String, Integer> sortMessageKeyByIds(Map<String, Integer> messageKeyIds) {

		Comparator<String> cmp = new MessageIdComparator(messageKeyIds);
		Map<String, Integer> sortedMessageKeyIds = new TreeMap<>(cmp);
		sortedMessageKeyIds.putAll(messageKeyIds);
		return sortedMessageKeyIds;
	}

	public void generate() throws IOException {
		generateNLSBinaryAndImmutableFile();
		generateNLSInterface();
		generateImmutableListFile();
		generateNLSSystemPropertiesFile();
		gerenateNLSResourcesListFile();
	}

	private void generateNLSInterface() throws IOException {
		// generate NLS interface
		String nlsInterfaceName = this.header;
		if (this.interfaceNameToUpperCase) {
			nlsInterfaceName = nlsInterfaceName.toUpperCase();
		}
		ST nlsInterface = NlsSTGroup.getInstanceOf("NLSInterface"); //$NON-NLS-1$
		nlsInterface.add("package", this.packageName); //$NON-NLS-1$
		nlsInterface.add("interfaceName", nlsInterfaceName); //$NON-NLS-1$
		for (Entry<String, Integer> entry : this.messageKeyIds.entrySet()) {
			String msgKey = entry.getKey();
			int msgId = entry.getValue().intValue();
			nlsInterface.addAggr("messageKeyIds.{msgKey, msgId}", msgKey, new Integer(msgId)); //$NON-NLS-1$
		}
		File nlsInterfaceFile = new File(this.nlsGeneratedDir, nlsInterfaceName + JAVA);
		writeGeneratedCode(nlsInterface, nlsInterfaceFile);
	}

	private void generateNLSBinaryAndImmutableFile() throws IOException {
		this.out = new WatchableByteArrayOutputStream();
		this.positions = new HashMap<String, Integer>();
		ST nlsImmutableST = NlsSTGroup.getInstanceOf("NLSImmutable"); //$NON-NLS-1$

		// locale names immutable string template
		ST nlsLocaleNamesImmutableST = NlsSTGroup.getInstanceOf("NLSLocale"); //$NON-NLS-1$
		nlsLocaleNamesImmutableST.add("name", this.header + "_locales"); //$NON-NLS-1$ //$NON-NLS-2$
		nlsLocaleNamesImmutableST.add("entryCount", new Integer(this.locales.length)); //$NON-NLS-1$

		for (NLSLocale locale : this.locales) {

			// write the locale name and add its position
			int position = writeString(locale.getLocaleName());
			nlsLocaleNamesImmutableST.add("positions", new Integer(position)); //$NON-NLS-1$

			// locale immutable string template
			ST nlsLocaleImmutableST = NlsSTGroup.getInstanceOf("NLSLocale"); //$NON-NLS-1$

			// add the locale immutable name
			String immutableLocaleName = this.header + "_" + locale.getLocaleName(); //$NON-NLS-1$
			nlsLocaleImmutableST.add("name", immutableLocaleName); //$NON-NLS-1$

			// write the display name and add its position
			position = writeString(locale.getDisplayName());
			nlsLocaleImmutableST.add("positions", new Integer(position)); //$NON-NLS-1$

			int localeTranslationCount = getMaxId();
			// process locale translations
			for (int id = 0; ++id <= localeTranslationCount;) {
				position = -1;
				Integer msgId = new Integer(id);
				if (this.messageKeyIds.containsValue(msgId)) {
					String msgKey = getMessageKeyById(msgId);
					String translation = locale.getTranslation(msgKey);
					if (translation != null) {
						position = writeString(translation);
					}
				}
				nlsLocaleImmutableST.add("positions", new Integer(position)); //$NON-NLS-1$
			}

			// +1 for display name
			nlsLocaleImmutableST.add("entryCount", new Integer(localeTranslationCount + 1)); //$NON-NLS-1$

			nlsImmutableST.add("locales", nlsLocaleImmutableST); //$NON-NLS-1$
		}
		nlsImmutableST.add("localeNames", nlsLocaleNamesImmutableST); //$NON-NLS-1$
		// generate NLS immutable file
		File nlsImmutableFile = new File(this.nlsGeneratedDir, this.header + XML);
		writeGeneratedCode(nlsImmutableST, nlsImmutableFile);

		// generate the NLS binary file
		File nlsBinaryFile = new File(this.nlsGeneratedDir, this.header + BIN);
		try (FileOutputStream binaryFileOut = new FileOutputStream(nlsBinaryFile)) {
			this.out.flush();
			this.out.writeTo(binaryFileOut);
			this.out.close();
		}

	}

	// write a string to the outputStream if not already written and return its
	// position.
	private int writeString(String string) {
		Integer position = this.positions.get(string);
		if (position != null) {
			// string already written, return its position
			return position.intValue();
		}
		// new string
		byte[] stringBytes = this.stringEncoder.encodeString(string);
		int pos = this.out.getPosition();
		this.out.write(stringBytes, 0, stringBytes.length);
		this.positions.put(string, new Integer(pos));
		return pos;
	}

	/*
	 * generate xxx.immutables.list file
	 */
	private void generateImmutableListFile() throws IOException {
		ST immutablesListTemplate = NlsSTGroup.getInstanceOf("List"); //$NON-NLS-1$
		immutablesListTemplate.add("entries", this.packageDir + this.header + XML); //$NON-NLS-1$
		File resourcesListFile = new File(this.nlsGeneratedPropertiesDir, this.header + IMMUTABLES_LIST);
		writeGeneratedCode(immutablesListTemplate, resourcesListFile);
	}

	/*
	 * generate xxx.system.properties file
	 */
	private void generateNLSSystemPropertiesFile() throws IOException {
		Map<String, String> properties = new HashMap<String, String>();
		// NLS header property
		properties.put(this.header + HEADER, this.header);
		// NLS binary file property
		properties.put(this.header + BINARY_FILE, this.packageDir + this.header + BIN);

		ST propertyTemplate = NlsSTGroup.getInstanceOf("Properties"); //$NON-NLS-1$
		propertyTemplate.add("properties", properties); //$NON-NLS-1$
		File propertyFile = new File(this.nlsGeneratedPropertiesDir, this.header + SYSTEM_PROPERTIES);
		writeGeneratedCode(propertyTemplate, propertyFile);
	}

	/*
	 * generate xxx.resources.list
	 */
	private void gerenateNLSResourcesListFile() throws IOException {
		ST resourcesListTemplate = NlsSTGroup.getInstanceOf("List"); //$NON-NLS-1$
		resourcesListTemplate.add("entries", this.packageDir + this.header + BIN); //$NON-NLS-1$
		File resourcesListFile = new File(this.nlsGeneratedPropertiesDir, this.header + RESOURCES_LIST);
		writeGeneratedCode(resourcesListTemplate, resourcesListFile);
	}

	static void writeGeneratedCode(ST factoryTemplate, File file) throws IOException {
		try (BufferedWriter out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), "UTF-8"))) { //$NON-NLS-1$
			out.write(factoryTemplate.render());
			out.flush();
		}
	}

	static void createOutputGeneratedDir(File output) {
		if (output.isDirectory()) {
			return;
		}
		if (!output.mkdirs()) {
			throw new RuntimeException(
					String.format("Can not create the output directory: %s", output.getAbsolutePath()));
		}
	}

	class MessageIdComparator implements Comparator<String> {
		Map<String, Integer> map = new HashMap<String, Integer>();

		MessageIdComparator(Map<String, Integer> messageKeyIds) {
			this.map.putAll(messageKeyIds);
		}

		@Override
		public int compare(String msgKey1, String msgKey2) {
			if (this.map.get(msgKey1).intValue() >= this.map.get(msgKey2).intValue()) {
				return 1;
			}
			return -1;
		}
	}

}
