/*
 * Copyright (c) 2015, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2017, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security;

import java.io.IOException;
import java.io.InputStream;
import java.security.Key;
import java.security.KeyStoreException;
import java.security.KeyStoreSpi;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.util.Enumeration;
import java.util.Hashtable;

import com.is2t.support.security.x509.X509KeyImpl;

/**
 * This default keystore implementation supports both certificate and key entry.
 * <p>
 * The content of this keystore can not be load from an inputstream when the
 * method {@link #engineLoad(InputStream, char[])} is called.
 * <p>
 * The methods {@link #engineSetCertificateEntry(String, Certificate)} and
 * {@link #engineDeleteEntry(String)} are used to add/remove certificates
 * to/from this keystore. The method
 * {@link #engineSetKeyEntry(String, byte[], Certificate[])} is used to set a
 * private key with it associated certificate chain. the private key must be an
 * encrypted DER (ASN1) private key in PKCS#8 format.
 *
 */
public class DefaultKeyStore extends KeyStoreSpi {

	// Private keys and their supporting certificate chains
	private static class KeyEntry {
		Key privKey;
		Certificate chain[];
	};

	private final Hashtable<String, Object> entries;

	public DefaultKeyStore() {
		entries = new Hashtable<String, Object>();
	}

	@Override
	public Key engineGetKey(String alias, char[] password) throws NoSuchAlgorithmException, UnrecoverableKeyException {
		Object entry = entries.get(alias);
		if (entry == null || !(entry instanceof KeyEntry)) {
			return null;
		}
		if (password == null) {
			throw new UnrecoverableKeyException("Password must not be null");
		}
		return ((KeyEntry) entry).privKey;
	}

	@Override
	public Certificate[] engineGetCertificateChain(String alias) {
		Object entry = entries.get(alias);
		if (entry != null && entry instanceof KeyEntry) {
			if (((KeyEntry) entry).chain == null) {
				return null;
			} else {
				return ((KeyEntry) entry).chain.clone();
			}
		} else {
			return null;
		}
	}

	@Override
	public Certificate engineGetCertificate(String alias) {
		Object entry = entries.get(alias);
		if (entry != null) {
			if (entry instanceof Certificate) {
				return (Certificate) entry;
			} else {
				if (((KeyEntry) entry).chain == null) {
					return null;
				} else {
					return ((KeyEntry) entry).chain[0];
				}
			}
		} else {
			return null;
		}
	}

	@Override
	public void engineSetKeyEntry(String alias, byte[] key, Certificate[] chain) throws KeyStoreException {
		synchronized (entries) {
			KeyEntry entry = new KeyEntry();
			entry.privKey = new X509KeyImpl(key.clone());
			if ((chain != null) && (chain.length != 0)) {
				entry.chain = chain.clone();
			} else {
				entry.chain = null;
			}
			entries.put(alias, entry);
		}
	}

	@Override
	public void engineSetCertificateEntry(String alias, Certificate cert) throws KeyStoreException {
		synchronized (entries) {
			Object entry = entries.get(alias);
			if ((entry != null) && (entry instanceof KeyEntry)) {
				throw new KeyStoreException("Cannot overwrite own certificate");
			}
			entries.put(alias, cert);
		}
	}

	@Override
	public void engineDeleteEntry(String alias) throws KeyStoreException {
		synchronized (entries) {
			entries.remove(alias);
		}

	}

	@Override
	public Enumeration<String> engineAliases() {
		return entries.keys();
	}

	@Override
	public boolean engineContainsAlias(String alias) {
		return entries.contains(alias);
	}

	@Override
	public int engineSize() {
		return entries.size();
	}

	@Override
	public boolean engineIsKeyEntry(String alias) {
		Object entry = entries.get(alias);
		if ((entry != null) && (entry instanceof KeyEntry)) {
			return true;
		} else {
			return false;
		}
	}

	@Override
	public boolean engineIsCertificateEntry(String alias) {
		Object entry = entries.get(alias);
		if ((entry != null) && (entry instanceof Certificate)) {
			return true;
		} else {
			return false;
		}
	}

	@Override
	public void engineLoad(InputStream stream, char[] password)
			throws IOException, NoSuchAlgorithmException, CertificateException {

		// nothing to do
		// this keystore implementation does not support keystore loading from
		// an inputstream
	}

}
