/*
 * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2017-2019, MicroEJ Corp. - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */
package com.is2t.support.security.cipher;

import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.ProviderException;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;

import javax.crypto.AEADBadTagException;
import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherSpi;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.JceSecurity;
import javax.crypto.ShortBufferException;
import javax.crypto.spec.IvParameterSpec;

/**
 * Implementation copied from <code>com.sun.crypto.provider.CipherCore</code>.
 */
abstract public class AbstractCipherSpi extends CipherSpi {

	/**
	 * Originally defined in com.sun.crypto.provider.GaloisCounterMode.
	 */
	private static int GaloisCounterMode_DEFAULT_IV_LEN = 12; // in bytes

	/*
	 * block size of cipher in bytes
	 */
	private final int blockSize;
	/*
	 * unit size (number of input bytes that can be processed at a time)
	 */
	private final int unitBytes;

	/*
	 * index of the content size left in the buffer
	 */
	private int buffered = 0;

	private final int bufferLength;

	/**
	 * Lazy initialized buffer.
	 */
	private byte[] internalBuffer;

	/*
	 * minimum number of bytes in the buffer required for
	 * FeedbackCipher.encryptFinal()/decryptFinal() call.
	 * update() must buffer this many bytes before starting
	 * to encrypt/decrypt data.
	 * currently, only the following cases have non-zero values:
	 * 1) CTS mode - due to its special handling on the last two blocks
	 * (the last one may be incomplete).
	 * 2) GCM mode + decryption - due to its trailing tag bytes
	 */
	private final int minBytes = 0;

	/*
	 * number of bytes needed to make the total input length a multiple
	 * of the blocksize (this is used in feedback mode, when the number of
	 * input bytes that are processed at a time is different from the block
	 * size)
	 */
	private int diffBlocksize;

	/*
	 * TODO
	 * padding class
	 */
	private final Object padding = null;

	/*
	 * the cipher mode
	 */
	private final int cipherMode;

	/*
	 * are we encrypting or decrypting?
	 */
	private boolean decrypting = false;

	/*
	 * Block Mode constants.
	 * Native known.
	 */
	private static final int ECB_MODE = 0;
	private static final int CBC_MODE = 1;
	private static final int CFB_MODE = 2;
	private static final int OFB_MODE = 3;
	private static final int PCBC_MODE = 4;
	private static final int CTR_MODE = 5;
	private static final int CTS_MODE = 6;
	static final int GCM_MODE = 7;


	public AbstractCipherSpi(int blockSize) {
		this(blockSize, blockSize, ECB_MODE);
	}

	public AbstractCipherSpi(int blockSize, int unitBytes, int cipherMode) {
		this.blockSize = blockSize;
		this.diffBlocksize = blockSize;
		this.unitBytes = unitBytes;
		this.cipherMode = cipherMode;

		/*
		 * The buffer should be usable for all cipher mode and padding
		 * schemes. Thus, it has to be at least (blockSize+1) for CTS.
		 * In decryption mode, it also hold the possible padding block.
		 */
		this.bufferLength = blockSize * 2;
	}

	@Override
	protected int engineGetBlockSize() {
		return blockSize;
	}

	@Override
	protected int engineGetOutputSize(int inputLen) {
		return getOutputSizeByOperation(inputLen, true);
	}

	@Override
	protected void engineInit(int opmode, Key key, SecureRandom random) throws InvalidKeyException {
		try {
			engineInit(opmode, key, null, random);
		} catch (InvalidAlgorithmParameterException e) {
			e.printStackTrace();
		}
	}

	@Override
	protected void engineInit(int opmode, Key key, AlgorithmParameterSpec params, SecureRandom random) throws InvalidKeyException, InvalidAlgorithmParameterException {
		decrypting = (opmode == Cipher.DECRYPT_MODE)
				|| (opmode == Cipher.UNWRAP_MODE);

		byte[] keyBytes = getKeyBytes(key);
		int tagLen = -1;
		byte[] ivBytes = null;
		if (params != null) {
			if (cipherMode == GCM_MODE) {
				//				if (params instanceof GCMParameterSpec) {
				//					tagLen = ((GCMParameterSpec)params).getTLen();
				//					if (tagLen < 96 || tagLen > 128 || ((tagLen & 0x07) != 0)) {
				//						throw new InvalidAlgorithmParameterException
				//						("Unsupported TLen value; must be one of " +
				//								"{128, 120, 112, 104, 96}");
				//					}
				//					tagLen = tagLen >> 3;
				//					ivBytes = ((GCMParameterSpec)params).getIV();
				//				} else {
				throw new InvalidAlgorithmParameterException
				("Unsupported parameter: " + params);
				//				}
			} else {
				if (params instanceof IvParameterSpec) {
					ivBytes = ((IvParameterSpec)params).getIV();
					if ((ivBytes == null) || (ivBytes.length != blockSize)) {
						throw new InvalidAlgorithmParameterException
						("Wrong IV length: must be " + blockSize +
								" bytes long");
					}
				}
				//				else if (params instanceof RC2ParameterSpec) {
				//					ivBytes = ((RC2ParameterSpec)params).getIV();
				//					if ((ivBytes != null) && (ivBytes.length != blockSize)) {
				//						throw new InvalidAlgorithmParameterException
				//						("Wrong IV length: must be " + blockSize +
				//								" bytes long");
				//					}
				//				}
				else {
					throw new InvalidAlgorithmParameterException
					("Unsupported parameter: " + params);
				}
			}
		}
		if (cipherMode == ECB_MODE) {
			if (ivBytes != null) {
				throw new InvalidAlgorithmParameterException
				("ECB mode cannot use IV");
			}
		} else if (ivBytes == null)  {
			if (decrypting) {
				throw new InvalidAlgorithmParameterException("Parameters "
						+ "missing");
			}

			if (random == null) {
				random = JceSecurity.RANDOM;
			}
			if (cipherMode == GCM_MODE) {
				ivBytes = new byte[GaloisCounterMode_DEFAULT_IV_LEN];
			} else {
				ivBytes = new byte[blockSize];
			}
			random.nextBytes(ivBytes);
		}

		buffered = 0;
		diffBlocksize = blockSize;

		String algorithm = key.getAlgorithm();

		//		// GCM mode needs additional handling
		//		if (cipherMode == GCM_MODE) {
		//			if(tagLen == -1) {
		//				tagLen = GaloisCounterMode.DEFAULT_TAG_LEN;
		//			}
		//			if (decrypting) {
		//				minBytes = tagLen;
		//			} else {
		//				// check key+iv for encryption in GCM mode
		//				requireReinit =
		//						Arrays.equals(ivBytes, lastEncIv) &&
		//						MessageDigest.isEqual(keyBytes, lastEncKey);
		//				if (requireReinit) {
		//					throw new InvalidAlgorithmParameterException
		//					("Cannot reuse iv for GCM encryption");
		//				}
		//				lastEncIv = ivBytes;
		//				lastEncKey = keyBytes;
		//			}
		//			((GaloisCounterMode) cipher).init
		//			(decrypting, algorithm, keyBytes, ivBytes, tagLen);
		//		} else {
		engineInit(decrypting, algorithm, keyBytes, ivBytes);
		//		}
		// skip checking key+iv from now on until after doFinal()
		//		requireReinit = false;

	}

	/**
	 * Return the key bytes of the specified key. Throw an InvalidKeyException
	 * if the key is not usable.
	 */
	static byte[] getKeyBytes(Key key) throws InvalidKeyException {
		if (key == null) {
			throw new InvalidKeyException("No key given");
		}
		// note: key.getFormat() may return null
		if (!"RAW".equalsIgnoreCase(key.getFormat())) {
			throw new InvalidKeyException("Wrong format: RAW bytes needed");
		}
		byte[] keyBytes = key.getEncoded();
		if (keyBytes == null) {
			throw new InvalidKeyException("RAW key bytes missing");
		}
		return keyBytes;
	}

	@Override
	protected byte[] engineUpdate(byte[] input, int inputOffset, int inputLen) {
		byte[] output = null;
		try {
			output = new byte[getOutputSizeByOperation(inputLen, false)];
			int len = engineUpdate(input, inputOffset, inputLen, output,
					0);
			if (len == output.length) {
				return output;
			} else {
				byte[] packedOutput = new byte[len];
				System.arraycopy(output, 0, packedOutput, 0, len);
				return packedOutput;
			}
		} catch (ShortBufferException e) {
			// should never happen
			throw new ProviderException("Unexpected exception", e);
		}
	}

	@Override
	protected int engineUpdate(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset) throws ShortBufferException {
		// figure out how much can be sent to crypto function
		int len = addExact(buffered, inputLen);
		len -= minBytes;
		if (padding != null && decrypting) {
			// do not include the padding bytes when decrypting
			len -= blockSize;
		}
		// do not count the trailing bytes which do not make up a unit
		len = (len > 0 ? (len - (len % unitBytes)) : 0);

		// check output buffer capacity
		if ((output == null) ||
				((output.length - outputOffset) < len)) {
			throw new ShortBufferException("Output buffer must be "
					+ "(at least) " + len
					+ " bytes long");
		}

		int outLen = 0;
		if (len != 0) { // there is some work to do
			if ((input == output)
					&& (outputOffset - inputOffset < inputLen)
					&& (inputOffset - outputOffset < bufferLength)) {
				// copy 'input' out to avoid its content being
				// overwritten prematurely.
				byte[] newInput = new byte[inputLen];
				System.arraycopy(input, inputOffset, newInput, 0, inputLen);
				inputOffset = 0;
			}
			if (len <= buffered) {
				// all to-be-processed data are from 'buffer'
				byte[] buffer = getBuffer();
				if (decrypting) {
					outLen = decrypt(buffer, 0, len, output, outputOffset);
				} else {
					outLen = encrypt(buffer, 0, len, output, outputOffset);
				}
				buffered -= len;
				if (buffered != 0) {
					System.arraycopy(buffer, len, buffer, 0, buffered);
				}
			} else { // len > buffered
				int inputConsumed = len - buffered;
				int temp;
				if (buffered > 0) {
					int bufferCapacity = bufferLength - buffered;
					byte[] buffer = getBuffer();
					if (bufferCapacity != 0) {
						temp = Math.min(bufferCapacity, inputConsumed);
						if (unitBytes != blockSize) {
							temp -= (addExact(buffered, temp) % unitBytes);
						}
						System.arraycopy(input, inputOffset, buffer, buffered, temp);
						inputOffset = addExact(inputOffset, temp);
						inputConsumed -= temp;
						inputLen -= temp;
						buffered = addExact(buffered, temp);
					}
					// process 'buffer'
					if (decrypting) {
						outLen = decrypt(buffer, 0, buffered, output, outputOffset);
					} else {
						outLen = encrypt(buffer, 0, buffered, output, outputOffset);
					}
					outputOffset = addExact(outputOffset, outLen);
					buffered = 0;
				}
				if (inputConsumed > 0) { // still has input to process
					if (decrypting) {
						outLen += decrypt(input, inputOffset, inputConsumed,
								output, outputOffset);
					} else {
						outLen += encrypt(input, inputOffset, inputConsumed,
								output, outputOffset);
					}
					inputOffset += inputConsumed;
					inputLen -= inputConsumed;
				}
			}
			// Let's keep track of how many bytes are needed to make
			// the total input length a multiple of blocksize when
			// padding is applied
			if (unitBytes != blockSize) {
				if (len < diffBlocksize) {
					diffBlocksize -= len;
				} else {
					diffBlocksize = blockSize -
							((len - diffBlocksize) % blockSize);
				}
			}
		}
		// Store remaining input into 'buffer' again
		if (inputLen > 0) {
			byte[] buffer = getBuffer();
			System.arraycopy(input, inputOffset, buffer, buffered,
					inputLen);
			buffered = addExact(buffered, inputLen);
		}
		return outLen;
	}

	@Override
	protected byte[] engineDoFinal(byte[] input, int inputOffset, int inputLen) throws IllegalBlockSizeException, BadPaddingException {

		byte[] output = null;
		try {
			output = new byte[getOutputSizeByOperation(inputLen, true)];
			int len = engineDoFinal(input, inputOffset, inputLen, output, 0);
			if (len < output.length) {
				byte[] packedOutput = new byte[len];
				System.arraycopy(output, 0, packedOutput, 0, len);
				return packedOutput;
			} else {
				return output;
			}
		} catch (ShortBufferException e) {
			// never thrown
			throw new ProviderException("Unexpected exception", e);
		}
	}

	@Override
	protected int engineDoFinal(byte[] input, int inputOffset, int inputLen, byte[] output, int outputOffset) throws ShortBufferException, IllegalBlockSizeException, BadPaddingException {

		int estOutSize = getOutputSizeByOperation(inputLen, true);
		// check output buffer capacity.
		// if we are decrypting with padding applied, we can perform this
		// check only after we have determined how many padding bytes there
		// are.
		int outputCapacity = output.length - outputOffset;
		int minOutSize = (decrypting? (estOutSize - blockSize):estOutSize);
		if ((output == null) || (outputCapacity < minOutSize)) {
			throw new ShortBufferException("Output buffer must be "
					+ "(at least) " + minOutSize + " bytes long");
		}

		// calculate total input length
		int len = addExact(buffered, inputLen);

		// calculate padding length
		//TODO: cipher.getBufferedLength() is zero by default
		//		int totalLen = Math.addExact(len, cipher.getBufferedLength());
		int totalLen = len;
		int paddingLen = 0;
		// will the total input length be a multiple of blockSize?
		if (unitBytes != blockSize) {
			if (totalLen < diffBlocksize) {
				paddingLen = diffBlocksize - totalLen;
			} else {
				paddingLen = blockSize -
						((totalLen - diffBlocksize) % blockSize);
			}
		}
		else if (padding != null) {
			// TODO
			//paddingLen = padding.padLength(totalLen);
			throw new UnsupportedOperationException();
		}

		if (decrypting && (padding != null) &&
				(paddingLen > 0) && (paddingLen != blockSize)) {
			throw new IllegalBlockSizeException
			("Input length must be multiple of " + blockSize +
					" when decrypting with padded cipher");
		}

		/*
		 * prepare the final input, assemble a new buffer if any
		 * of the following is true:
		 *  - 'input' and 'output' are the same buffer
		 *  - there are internally buffered bytes
		 *  - doing encryption and padding is needed
		 */
		byte[] finalBuf = input;
		int finalOffset = inputOffset;
		int finalBufLen = inputLen;
		if ((buffered != 0) || (!decrypting && padding != null) ||
				((input == output)
						&& (outputOffset - inputOffset < inputLen)
						&& (inputOffset - outputOffset < bufferLength))) {
			if (decrypting || padding == null) {
				paddingLen = 0;
			}
			finalBuf = new byte[addExact(len, paddingLen)];
			finalOffset = 0;
			if (buffered != 0) {
				byte[] buffer = getBuffer();
				System.arraycopy(buffer, 0, finalBuf, 0, buffered);
			}
			if (inputLen != 0) {
				System.arraycopy(input, inputOffset, finalBuf,
						buffered, inputLen);
			}
			if (paddingLen != 0) {
				// TODO
				throw new UnsupportedOperationException();
				//				padding.padWithLen(finalBuf, addExact(buffered, inputLen), paddingLen);
			}
			finalBufLen = finalBuf.length;
		}
		int outLen = 0;
		if (decrypting) {
			// if the size of specified output buffer is less than
			// the length of the cipher text, then the current
			// content of cipher has to be preserved in order for
			// users to retry the call with a larger buffer in the
			// case of ShortBufferException.
			if (outputCapacity < estOutSize) {
				save();
			}
			// create temporary output buffer so that only "real"
			// data bytes are passed to user's output buffer.
			byte[] outWithPadding = new byte[estOutSize];
			outLen = finalNoPadding(finalBuf, finalOffset, outWithPadding,
					0, finalBufLen);

			if (padding != null) {
				// TODO
				throw new UnsupportedOperationException();

				//				int padStart = padding.unpad(outWithPadding, 0, outLen);
				//				if (padStart < 0) {
				//					throw new BadPaddingException("Given final block not "
				//							+ "properly padded");
				//				}
				//				outLen = padStart;
			}

			if (outputCapacity < outLen) {
				// restore so users can retry with a larger buffer
				restore();
				throw new ShortBufferException("Output buffer too short: "
						+ (outputCapacity)
						+ " bytes given, " + outLen
						+ " bytes needed");
			}
			// copy the result into user-supplied output buffer
			System.arraycopy(outWithPadding, 0, output, outputOffset, outLen);
		} else { // encrypting
			try {
				outLen = finalNoPadding(finalBuf, finalOffset, output,
						outputOffset, finalBufLen);
			} finally {
				// TODO
				// reset after doFinal() for GCM encryption
				//requireReinit = (cipherMode == GCM_MODE);
			}
		}

		buffered = 0;
		diffBlocksize = blockSize;
		if (cipherMode != ECB_MODE) {
			reset();
		}
		return outLen;
	}

	private int finalNoPadding(byte[] in, int inOfs, byte[] out, int outOfs,
			int len)
					throws IllegalBlockSizeException, AEADBadTagException,
					ShortBufferException {

		if ((cipherMode != GCM_MODE) && (in == null || len == 0)) {
			return 0;
		}
		if ((cipherMode != CFB_MODE) && (cipherMode != OFB_MODE) &&
				(cipherMode != GCM_MODE) &&
				((len % unitBytes) != 0) && (cipherMode != CTS_MODE)) {
			if (padding != null) {
				throw new IllegalBlockSizeException
				("Input length (with padding) not multiple of " +
						unitBytes + " bytes");
			} else {
				throw new IllegalBlockSizeException
				("Input length not multiple of " + unitBytes
						+ " bytes");
			}
		}
		int outLen = 0;
		if (decrypting) {
			outLen = decryptFinal(in, inOfs, len, out, outOfs);
		} else {
			outLen = encryptFinal(in, inOfs, len, out, outOfs);
		}
		return outLen;
	}


	/**
	 * Performs decryption operation for the last time.
	 *
	 * <p>NOTE: For cipher feedback modes which does not perform
	 * special handling for the last few blocks, this is essentially
	 * the same as <code>encrypt(...)</code>. Given most modes do
	 * not do special handling, the default impl for this method is
	 * to simply call <code>decrypt(...)</code>.
	 *
	 * @param cipher the input buffer with the data to be decrypted
	 * @param cipherOffset the offset in <code>cipher</code>
	 * @param cipherLen the length of the input data
	 * @param plain the buffer for the decryption result
	 * @param plainOffset the offset in <code>plain</code>
	 * @return the number of bytes placed into <code>plain</code>
	 */
	int decryptFinal(byte[] cipher, int cipherOffset, int cipherLen,
			byte[] plain, int plainOffset)
					throws IllegalBlockSizeException, AEADBadTagException,
					ShortBufferException {
		return decrypt(cipher, cipherOffset, cipherLen, plain, plainOffset);
	}

	/**
	 * Performs encryption operation for the last time.
	 *
	 * <p>NOTE: For cipher feedback modes which does not perform
	 * special handling for the last few blocks, this is essentially
	 * the same as <code>encrypt(...)</code>. Given most modes do
	 * not do special handling, the default impl for this method is
	 * to simply call <code>encrypt(...)</code>.
	 *
	 * @param plain the input buffer with the data to be encrypted
	 * @param plainOffset the offset in <code>plain</code>
	 * @param plainLen the length of the input data
	 * @param cipher the buffer for the encryption result
	 * @param cipherOffset the offset in <code>cipher</code>
	 * @return the number of bytes placed into <code>cipher</code>
	 */
	int encryptFinal(byte[] plain, int plainOffset, int plainLen,
			byte[] cipher, int cipherOffset)
					throws IllegalBlockSizeException, ShortBufferException {
		return encrypt(plain, plainOffset, plainLen, cipher, cipherOffset);
	}

	/**
	 * Returns the sum of its arguments,
	 * throwing an exception if the result overflows an {@code int}.
	 *
	 * @param x the first value
	 * @param y the second value
	 * @return the result
	 * @throws ArithmeticException if the result overflows an int
	 * @since 1.8
	 */
	public static int addExact(int x, int y) {
		int r = x + y;
		// HD 2-12 Overflow iff both arguments have the opposite sign of the result
		if (((x ^ r) & (y ^ r)) < 0) {
			throw new ArithmeticException("integer overflow");
		}
		return r;
	}

	private byte[] getBuffer() {
		if(internalBuffer == null) {
			this.internalBuffer = new byte[bufferLength];
		}
		return internalBuffer;
	}


	/**
	 * Resets the iv to its original value.
	 * This is used when doFinal is called in the Cipher class, so that the
	 * cipher can be reused (with its original iv).
	 */
	private void reset() {
		// TODO

	}

	/**
	 * Restores the content of this cipher to the previous saved one.
	 */
	private void restore() {
		// TODO

	}

	/**
	 * Save the current content of this cipher.
	 */
	private void save() {
		// TODO

	}

	// TODO: here algorithm is retrieved from the Key. How to check if this is ok ?
	/**
	 * Initializes the cipher in the specified mode with the given key
	 * and iv.
	 *
	 * @param decrypting flag indicating encryption or decryption mode
	 * @param algorithm the algorithm name (never null)
	 * @param key the key (never null)
	 * @param iv the iv (either null or blockSize bytes long)
	 *
	 * @exception InvalidKeyException if the given key is inappropriate for
	 * initializing this cipher
	 */
	abstract protected void engineInit(boolean decrypting, String algorithm, byte[] key,
			byte[] iv) throws InvalidKeyException;

	// TODO: see if we use the OpenJDK Implementation in com.sun.crypto.provider.CipherCore.getOutputSizeByOperation(int, boolean).
	//abstract protected int getOutputSizeByOperation(int inputLen, boolean isDoFinal);

	private int getOutputSizeByOperation(int inputLen, boolean isDoFinal) {
		int totalLen = addExact(buffered, getBufferedLength());
		totalLen = addExact(totalLen, inputLen);
		switch (cipherMode) {
			case GCM_MODE:
				// TODO
				throw new RuntimeException();
				//				if (isDoFinal) {
				//					int tagLen = ((GaloisCounterMode) cipher).getTagLen();
				//					if (!decrypting) {
				//						totalLen = Math.addExact(totalLen, tagLen);
				//					} else {
				//						totalLen -= tagLen;
				//					}
				//				}
				//				if (totalLen < 0) {
				//					totalLen = 0;
				//				}
				//				break;
			default:
				if (padding != null && !decrypting) {
					if (unitBytes != blockSize) {
						if (totalLen < diffBlocksize) {
							totalLen = diffBlocksize;
						} else {
							int residue = (totalLen - diffBlocksize) % blockSize;
							totalLen = addExact(totalLen, (blockSize - residue));
						}
					} else {
						// TODO padding
						throw new RuntimeException();
						//totalLen = Math.addExact(totalLen, padding.padLength(totalLen));
					}
				}
				break;
		}
		return totalLen;
	}



	/**
	 * TODO
	 *
	 * @return natively buffered length
	 */
	abstract protected int getBufferedLength();

	/**
	 * Performs encryption operation.
	 *
	 * <p>The input <code>plain</code>, starting at <code>plainOffset</code>
	 * and ending at <code>(plainOffset+plainLen-1)</code>, is encrypted.
	 * The result is stored in <code>cipher</code>, starting at
	 * <code>cipherOffset</code>.
	 *
	 * <p>The subclass that implements Cipher should ensure that
	 * <code>init</code> has been called before this method is called.
	 *
	 * @param plain the input buffer with the data to be encrypted
	 * @param plainOffset the offset in <code>plain</code>
	 * @param plainLen the length of the input data
	 * @param cipher the buffer for the encryption result
	 * @param cipherOffset the offset in <code>cipher</code>
	 * @return the number of bytes placed into <code>cipher</code>
	 */
	abstract protected int encrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset);

	/**
	 * Performs decryption operation.
	 *
	 * <p>The input <code>cipher</code>, starting at <code>cipherOffset</code>
	 * and ending at <code>(cipherOffset+cipherLen-1)</code>, is decrypted.
	 * The result is stored in <code>plain</code>, starting at
	 * <code>plainOffset</code>.
	 *
	 * <p>The subclass that implements Cipher should ensure that
	 * <code>init</code> has been called before this method is called.
	 *
	 * @param cipher the input buffer with the data to be decrypted
	 * @param cipherOffset the offset in <code>cipher</code>
	 * @param cipherLen the length of the input data
	 * @param plain the buffer for the decryption result
	 * @param plainOffset the offset in <code>plain</code>
	 * @return the number of bytes placed into <code>plain</code>
	 */
	abstract protected int decrypt(byte[] buffer, int i, int len, byte[] output, int outputOffset);

}
