/*
 * Java
 *
 * Copyright 2020-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.microej.vivante.utils;

import com.verisilicon.vivante.vg_lite_path;

/**
 * Utility base class for building {@link vg_lite_path}.<br>
 * See Section <code>2.4 Vector Path Data Opcodes</code> of <code>VGLite® Graphics API v2.07</code>.
 */
public abstract class PathBufferBase {
	/**
	 * Closes a path.
	 */
	protected static final int CMD_CLOSE = vg_lite_path.VG_LITE_CMD_CLOSE;

	/**
	 * Move to a destination.
	 */
	protected static final int CMD_MOVE = vg_lite_path.VG_LITE_CMD_MOVE;

	/**
	 * Line from current position to an absolute destination.
	 */
	protected static final int CMD_LINE = vg_lite_path.VG_LITE_CMD_LINE;

	/**
	 * Line from current position to a relative destination.
	 */
	protected static final int CMD_LINE_REL = vg_lite_path.VG_LITE_CMD_LINE_REL;

	/**
	 * Quadratic curve from current position to an absolute destination.
	 *
	 * Note: Requires one absolute control point
	 */
	protected static final int CMD_QUAD = vg_lite_path.VG_LITE_CMD_QUAD;

	/**
	 * Quadratic curve from current position to an relative destination.
	 *
	 * Note: Requires one absolute relative point
	 */
	protected static final int CMD_QUAD_REL = vg_lite_path.VG_LITE_CMD_QUAD_REL;

	/**
	 * Cubic curve from current position to an absolute destination.
	 *
	 * Note: Requires two absolute control points
	 */
	protected static final int CMD_CUBIC = vg_lite_path.VG_LITE_CMD_CUBIC;

	/**
	 * Cubic curve from current position to an relative destination.
	 *
	 * Note: Requires two absolute relative point
	 */
	protected static final int CMD_CUBIC_REL = vg_lite_path.VG_LITE_CMD_CUBIC_REL;

	private static final int CHUNK_SIZE = 20;

	int index;
	byte[] path;
	boolean closed;

	/**
	 * Constructor for the PathBufferBase class
	 *
	 * This constructor allocates a chunk of memory for the path data
	 */
	public PathBufferBase() {
		this.path = new byte[CHUNK_SIZE];
		this.index = 0;
		this.closed = false;
	}

	/**
	 * Extends the path data array with an extra chunk
	 *
	 * @return The new path data array
	 */
	protected byte[] extend() {
		byte[] previousPath = this.path;

		byte[] newPath = new byte[previousPath.length + CHUNK_SIZE];
		System.arraycopy(previousPath, 0, newPath, 0, this.index);
		this.path = newPath;
		return newPath;
	}

	/**
	 * Extends the path data array to contain a given number of fields
	 *
	 * @param nbFields
	 *            Number of fields that will be added
	 *
	 * @return the new path data array
	 */
	protected byte[] extend(int nbFields) {
		while (this.path.length - this.index < (nbFields * getUnitSize())) {
			this.path = extend();
		}

		return this.path;
	}

	/**
	 * @return The size of a field
	 */
	protected abstract int getUnitSize();

	/**
	 * @return the vg_lite path format
	 */
	public abstract int getFormat();

	/**
	 * Adds a command with zero points in the path data
	 *
	 * @param cmd
	 *            the command to add in the path data
	 */
	protected abstract void command(int cmd);

	/**
	 * @return true if the path is already closed, false otherwise
	 */
	public boolean isClosed() {
		return this.closed;
	}

	/**
	 * Closes the path
	 */
	public void close() {
		command(CMD_CLOSE);
		this.closed = true;
	}

	/**
	 * Resets the path
	 */
	public void reset() {
		this.index = 0;
		this.closed = false;
	}

	/**
	 * Convert the path to a {@link vg_lite_path}.
	 *
	 * @param left
	 *            left bound of the path
	 * @param top
	 *            top bound of the path
	 * @param right
	 *            right bound of the path
	 * @param bottom
	 *            bottom bound of the path
	 * @param quality
	 *            vg_lite quality of the path
	 * @return a new instance of the {@link vg_lite_path} class
	 */
	public vg_lite_path toVGLitePath(float left, float top, float right, float bottom, int quality) {
		if (!isClosed()) {
			close();
		}

		return new vg_lite_path(left, top, right, bottom, quality, getFormat(), this.path, 0, this.index);
	}

	/**
	 * Updates an existing {@link vg_lite_path}.
	 *
	 * @param vgLitePath
	 *            the path to update *
	 * @param left
	 *            left bound of the path
	 * @param top
	 *            top bound of the path
	 * @param right
	 *            right bound of the path
	 * @param bottom
	 *            bottom bound of the path
	 * @param quality
	 *            vg_lite quality of the path
	 */
	public void updateVGLitePath(vg_lite_path vgLitePath, float left, float top, float right, float bottom,
			int quality) {
		if (!isClosed()) {
			close();
		}

		vgLitePath.update(left, top, right, bottom, quality, getFormat(), this.path, 0, this.index);
	}
}
