/*
 * Java
 *
 * Copyright 2019-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package com.verisilicon.vivante;

import java.awt.Graphics2D;
import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.GeneralPath;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;

/**
 * Matrix is the class providing affine transformations.
 * <p>
 * A Matrix can be filled with multiple transformation operations. These operations can be identified by eiher
 * <ul>
 * <li>a Graphics2d object</li>
 * <li>a GeneralPath object</li>
 * <li>2 Point2D objects</li>
 * <li>a Shape object</li>
 * </ul>
 */
public class Matrix {
	private final List<MatrixOperation> ops;

	/**
	 * Create an empty Matrix.
	 */
	public Matrix() {
		this.ops = new ArrayList<>();
	}

	/**
	 * Get an handle to the matrix object.
	 *
	 * @return the hashCode of the matrix object.
	 **/
	public int getHandle() {
		return this.hashCode();
	}

	/**
	 * Add a matrix operation.
	 *
	 * @param op
	 *            the operation to add to the matrix.
	 **/
	private void add(MatrixOperation op) {
		this.ops.add(op);
	}

	private void clear() {
		this.ops.clear();
	}

	/**
	 * Apply all operations of the matrix to a Graphics2d object.
	 *
	 * @param g2d
	 *            Graphics2d object used to transform the matrix.
	 **/
	public void transform(Graphics2D g2d) {
		ListIterator<MatrixOperation> iter = this.ops.listIterator(this.ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(g2d);
		}
	}

	/**
	 * Transform the matrix with a GeneralPath object.
	 *
	 * @param p
	 *            GeneralPath object used to transform the matrix.
	 **/
	public void transform(GeneralPath p) {
		ListIterator<MatrixOperation> iter = this.ops.listIterator(this.ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(p);
		}
	}

	/**
	 * Transform the matrix with 2 Point2D objects.
	 *
	 * @param source
	 *            source point used to transform the matrix.
	 * @param destination
	 *            destination point used to transform the matrix.
	 **/
	public void transform(Point2D source, Point2D destination) {
		ListIterator<MatrixOperation> iter = this.ops.listIterator(this.ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			op.transform(source, destination);
			source = destination;
		}
	}

	/**
	 * Transform the matrix with a Shape object.
	 *
	 * @param shape
	 *            Shape object used to transform the matrix.
	 * @return shape object with previous existing operations.
	 **/
	public Shape transform(Shape shape) {
		ListIterator<MatrixOperation> iter = this.ops.listIterator(this.ops.size());

		while (iter.hasPrevious()) {
			MatrixOperation op = iter.previous();

			shape = op.transform(shape);
		}

		return shape;
	}

	/**
	 * Clear all operations and initialize the matrix as an identity matrix.
	 */
	public void identity() {
		// clear transformation
		clear();
		// apply a null translation in order to apply an identity transformation
		add(new OpTranslate(0, 0));
	}

	/**
	 * Add a translate operation to the list of operations.
	 *
	 * @param x
	 *            translation horizontal offset.
	 * @param y
	 *            translation vertical offset.
	 */
	public void translate(float x, float y) {
		add(new OpTranslate(x, y));
	}

	/**
	 *
	 * Add a scale operation to the list of operations.
	 *
	 * @param scale_x
	 *            horizontal scaling factor.
	 * @param scale_y
	 *            vertical scaling factor.
	 */
	public void scale(float scale_x, float scale_y) {
		add(new OpScale(scale_x, scale_y));
	}

	/**
	 *
	 * Add a rotate operation to the list of operations.
	 *
	 * @param degrees
	 *            rotation angle in degrees.
	 *
	 */
	public void rotate(float degrees) {
		add(new OpRotate(degrees));
	}

	/**
	 * Apply a perspective operation to a matrix. Unsupported
	 *
	 * @param px
	 *            w0 of perspective transformation matrix.
	 * @param py
	 *            w1 of perspective transformation matrix.
	 * @param matrix
	 *            matrix to apply the operation on.
	 *
	 */
	public static void perspective(float px, float py, Matrix matrix) {
		throw new UnsupportedOperationException();
	}

	/**
	 * Init matrix with an other matrix operations.
	 *
	 * @param srcMatrix
	 *            Source matrix
	 */
	public void copy(Matrix srcMatrix) {
		clear();

		for (MatrixOperation op : srcMatrix.ops) {
			this.add(op);
		}

	}

	private static interface MatrixOperation {
		public void transform(Graphics2D g2d);

		public void transform(GeneralPath p);

		public void transform(Point2D source, Point2D destination);

		public Shape transform(Shape shape);
	}

	private static class OpTranslate implements MatrixOperation {
		private final double _x;
		private final double _y;

		public OpTranslate(double x, double y) {
			this._x = x;
			this._y = y;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.translate(this._x, this._y);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.translate(this._x, this._y);
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.translate(this._x, this._y);
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.translate(this._x, this._y);
			return a.createTransformedShape(shape);
		}
	}

	private static class OpScale implements MatrixOperation {
		private final double _x;
		private final double _y;

		public OpScale(double x, double y) {
			this._x = x;
			this._y = y;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.scale(this._x, this._y);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.scale(this._x, this._y);
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.scale(this._x, this._y);
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.scale(this._x, this._y);
			return a.createTransformedShape(shape);
		}
	}

	private static class OpRotate implements MatrixOperation {
		private final double _t;

		public OpRotate(double t) {
			this._t = t;
		}

		@Override
		public void transform(Graphics2D g2d) {
			g2d.rotate(this._t);
		}

		@Override
		public void transform(GeneralPath p) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(this._t));
			p.transform(a);
		}

		@Override
		public void transform(Point2D source, Point2D destination) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(this._t));
			a.transform(source, destination);
		}

		@Override
		public Shape transform(Shape shape) {
			AffineTransform a = new AffineTransform();

			a.rotate(Math.toRadians(this._t));
			return a.createTransformedShape(shape);
		}
	}
}
