/*
 * Java
 *
 * Copyright 2016-2020 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.ecom.wifi;

import java.io.IOException;
import java.util.ArrayList;

import ej.bon.Immortals;
import ej.ecom.wifi.natives.GenericWifiManagerNatives;
import ej.ecom.wifi.natives.SecurityModeNativeConstants;
import ej.ecom.wifi.natives.WPSModeNativeConstants;
import ej.ecom.wifi.natives.WifiClientStateNativeConstants;
import ej.sni.SNI;

/**
 * An implementation of {@link WifiManager} module.
 */
public class WifiManager {

	private static final int J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT = -22;

	private static final int MAX_SSID_SIZE = 33; // max ssid size = 32 bytes + '\0'
	private static final int MAX_NAME_SIZE = 32; // max name size = 32 bytes
	private static final int MAX_BSSID_SIZE = 6; // max bssid size = 6 bytes (bssid == MAC address)
	private static final int MAX_RSSI_SIZE = 1; // max rssi size = 1 float
	private static final int MAX_WPA2_PASSPHRASE_SIZE = 64; // max WPA2 passphrase size = 64 bytes (8-63 ASCII characters)

	// Buffers and flags
	private byte[] ssid;
	private byte[] bssid;
	private float[] rssi;
	private byte[] passphraseBuff;
	private byte[] name;

	private boolean clientStarted;
	private boolean softAPStarted;

	// Unique instance for singleton
	private static final WifiManager INSTANCE = new WifiManager();

	// Private constructor for singleton
	private WifiManager() {
		Immortals.run(new Runnable() {

			@Override
			public void run() {
				ssid = new byte[MAX_SSID_SIZE];
				name = new byte[MAX_NAME_SIZE];
				bssid = new byte[MAX_BSSID_SIZE];
				rssi = new float[MAX_RSSI_SIZE];
				passphraseBuff = new byte[MAX_WPA2_PASSPHRASE_SIZE];
			}
		});
	}

	/**
	 * Get the unique {@link WifiManager} instance.
	 *
	 * @return the Wi-Fi manager
	 */
	public static WifiManager getInstance() {
		return INSTANCE;
	}

	public WifiCapability getCapability() throws IOException {
		int res = GenericWifiManagerNatives.getCapability(false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.getCapability(true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		return GenericWifiManagerNatives.convertNativeCapability(res);
	}

	public void join(AccessPoint accessPoint, String passphrase, SecurityMode securityMode) throws IOException {
		if (accessPoint == null) {
			throw new IllegalArgumentException();
		}

		join(accessPoint.getSSID(), passphrase, securityMode);
	}

	public void join(AccessPoint accessPoint, String passphrase) throws IOException {
		if (accessPoint == null) {
			throw new IllegalArgumentException();
		}

		join(accessPoint.getSSID(), passphrase, accessPoint.getSecurityMode());
	}

	public void join(String SSID, String passphrase) throws IOException {
		join(SSID, passphrase, null);
	}

	public void join(String SSID, String passphrase, SecurityMode securityMode) throws IOException {
		if ((getCapability() == WifiCapability.BOTH_EXCLUSIVE) && (this.softAPStarted)) {
			throw new IOException("Client and softAP modes cannot be used simultaneously");
		}

		synchronized (this) {
			byte[] ssidBuff = this.ssid;
			byte[] passphraseBuff = this.passphraseBuff;
			SNI.toCString(SSID, ssidBuff);
			SNI.toCString(passphrase, passphraseBuff);

			int err = 0;
			if(securityMode == null){
				err = GenericWifiManagerNatives.join(ssidBuff, 0, ssidBuff.length, passphraseBuff, 0, passphraseBuff.length,
						false);
			}else{
				err = GenericWifiManagerNatives.joinWithSecurityMode(ssidBuff, 0, ssidBuff.length, passphraseBuff, 0, passphraseBuff.length,
						SecurityModeNativeConstants.convertJavaSecurityModeToNativeMode(securityMode), false);
			}

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				if(securityMode == null){
					err = GenericWifiManagerNatives.join(ssidBuff, 0, ssidBuff.length, passphraseBuff, 0, passphraseBuff.length,
							true);
				}else{
					err = GenericWifiManagerNatives.joinWithSecurityMode(ssidBuff, 0, ssidBuff.length, passphraseBuff, 0, passphraseBuff.length,
							SecurityModeNativeConstants.convertJavaSecurityModeToNativeMode(securityMode), true);
				}
			}

			// check errors
			if(err < 0){
				throw new IOException("I/O error (err: "+err+")");
			}

			clientStarted = true;
		}

	}

	private void updateClientStartedState() throws IOException {
		int res = GenericWifiManagerNatives.getClientState(false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.getClientState(true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		clientStarted = res == WifiClientStateNativeConstants.STARTED;
	}

	public AccessPoint getJoined() throws IOException {
		AccessPoint ap = null;

		// Network may have been lost, update state
		if (clientStarted) {
			updateClientStartedState();
		} // else: we know the client is not started, hence we cannot lost network

		// Retrieve informations from native world if client is started
		if (clientStarted) {
			ap = new AccessPoint();

			synchronized (this) {
				// BSSID
				byte[] bssid = getBSSID();
				ap.setBSSID(bssid);

				// Channel
				int channel = getChannel();
				ap.setChannel(channel);

				// RSSI
				float rssi = getRSSI();
				ap.setRSSI(rssi);

				// Security mode
				SecurityMode securityMode = getSecurityMode();
				ap.setSecurityMode(securityMode);

				// SSID
				String ssid = getSSID();
				ap.setSSID(ssid);

				// WPS modes
				WPSMode[] wpsModes = getWPSModes();
				ap.setWPSModes(wpsModes);
			}

			// Network may have been lost while retrieving information, update state again
			updateClientStartedState();
			if (clientStarted == false) {
				// Network has been lost...
				ap = null;
			}
		}

		return ap;
	}

	private byte[] getBSSID() throws IOException {
		byte[] bssidBuff = this.bssid;
		int err = GenericWifiManagerNatives.getBSSID(bssidBuff, 0, bssidBuff.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = GenericWifiManagerNatives.getBSSID(bssidBuff, 0, bssidBuff.length, true);
		}

		// check errors
		if(err < 0){
			throw new IOException("I/O error (err: "+err+")");
		}

		byte[] bssid = new byte[MAX_BSSID_SIZE];
		System.arraycopy(bssidBuff, 0, bssid, 0, MAX_BSSID_SIZE);
		return bssid;
	}

	private int getChannel() throws IOException {
		int res = GenericWifiManagerNatives.getChannel(false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.getChannel(true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		return res;
	}

	private float getRSSI() throws IOException {
		float[] rssiBuff = this.rssi;
		int err = GenericWifiManagerNatives.getRSSI(rssiBuff, 0, rssiBuff.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = GenericWifiManagerNatives.getRSSI(rssiBuff, 0, rssiBuff.length, true);
		}

		// check errors
		if(err < 0){
			throw new IOException("I/O error (err: "+err+")");
		}
		return rssiBuff[0];
	}

	private SecurityMode getSecurityMode() throws IOException {
		int res = GenericWifiManagerNatives.getSecurityMode(false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.getSecurityMode(true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		return SecurityModeNativeConstants.convertNativeSecurityModeToJavaMode(res);
	}

	private String getSSID() throws IOException {
		byte[] ssidBuff = this.ssid;
		int err = GenericWifiManagerNatives.getSSID(ssidBuff, 0, ssidBuff.length, false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = GenericWifiManagerNatives.getSSID(ssidBuff, 0, ssidBuff.length, true);
		}

		// check errors
		if(err < 0){
			throw new IOException("I/O error (err: "+err+")");
		}

		return SNI.toJavaString(ssidBuff);
	}

	private WPSMode[] getWPSModes() throws IOException {
		int res = GenericWifiManagerNatives.getWPSModes(false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.getWPSModes(true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		return WPSModeNativeConstants.convertNativeWPSModeToJavaMode(res);
	}

	public void leave() throws IOException {
		int err = GenericWifiManagerNatives.leave(false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = GenericWifiManagerNatives.leave(true);
		}

		// check errors
		if(err < 0){
			throw new IOException("I/O error (err: "+err+")");
		}

		clientStarted = false;
	}

	public void enableSoftAP(SoftAPConfiguration configuration) throws IOException {
		if ((getCapability() == WifiCapability.BOTH_EXCLUSIVE) && (this.clientStarted)) {
			throw new IOException("Client and softAP modes cannot be used simultaneously");
		}

		// If a softap is already enabled, disable it before enabling the new one
		if (softAPStarted) {
			disableSoftAP();

		}

		synchronized (this) {
			SNI.toCString(configuration.getSSID(), this.ssid);
			String passphrase = configuration.getPassphrase();
			String name = configuration.getName();
			name = name == null ? "" : name; // avoid IllegalArgumentException in SNI.toCString()
			passphrase = passphrase == null ? "" : passphrase; // avoid IllegalArgumentException in SNI.toCString()
			SNI.toCString(passphrase, this.passphraseBuff);

			byte[] ssidBuff = this.ssid;
			byte[] passphraseBuff = this.passphraseBuff;

			/* Set Soft AP name */
			SNI.toCString(name, this.name);
			byte[] nameBuff = this.name;
			int err = GenericWifiManagerNatives.setNameSoftAP(nameBuff, 0,
					nameBuff.length, false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = GenericWifiManagerNatives.setNameSoftAP(nameBuff, 0,
						nameBuff.length, true);
			}

			// check errors
			if (err < 0) {
				throw new IOException("Cannot set softAP Name");
			}

			err = GenericWifiManagerNatives.enableSoftAP(ssidBuff, 0,
					ssidBuff.length, passphraseBuff, 0, passphraseBuff.length,
					false);

			// retry native returned with result
			while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
				err = GenericWifiManagerNatives.enableSoftAP(ssidBuff, 0,
						ssidBuff.length, passphraseBuff, 0,
						passphraseBuff.length, true);
			}

			// check errors
			if (err < 0) {
				throw new IOException("Cannot enable softAP");
			}
		}

		softAPStarted = true;
	}

	public void disableSoftAP() throws IOException {
		int err = GenericWifiManagerNatives.disableSoftAP(false);

		// retry native returned with result
		while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			err = GenericWifiManagerNatives.disableSoftAP(true);
		}

		// check errors
		if (err < 0) {
			throw new IOException("I/O error (err: "+err+")");
		}

		softAPStarted = false;
	}

	public boolean isSoftAPEnabled() throws IOException {
		return softAPStarted;
	}

	public AccessPoint[] scan(boolean active) throws IOException {
		int res = GenericWifiManagerNatives.scanAPCount(active, false);

		// retry native returned with result
		while (res == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
			res = GenericWifiManagerNatives.scanAPCount(active, true);
		}

		// check errors
		if(res < 0){
			throw new IOException("I/O error (err: "+res+")");
		}

		ArrayList<AccessPoint> aps = new ArrayList<AccessPoint>(0);
		int count = res;
		if(count > 0){
			int lastError = 0;
			for(int i = -1; ++i < count;){
				synchronized(this){
					/* Get SSID */
					int err = GenericWifiManagerNatives.scanAPssidAt(ssid, 0, ssid.length, i, false);

					// retry native returned with result
					while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						err = GenericWifiManagerNatives.scanAPssidAt(ssid, 0, ssid.length, i, true);
					}

					// check errors
					if(err < 0){
						lastError = err;
						continue;
					}

					/* Get BSSID */
					err = GenericWifiManagerNatives.scanAPbssidAt(bssid, 0, bssid.length, i, false);

					// retry native returned with result
					while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						err = GenericWifiManagerNatives.scanAPbssidAt(bssid, 0, bssid.length, i, true);
					}

					// check errors
					if(err < 0){
						lastError = err;
						continue;
					}

					/* Get Channel */
					int curAPChannel = GenericWifiManagerNatives.scanAPchannelAt(i, false);

					// retry native returned with result
					while (curAPChannel == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						curAPChannel = GenericWifiManagerNatives.scanAPchannelAt(i, true);
					}

					// check errors
					if (curAPChannel < 0) {
						lastError = curAPChannel;
						continue;
					}

					/* Get RSSI */
					err = GenericWifiManagerNatives.scanAPrssiAt(rssi, 0,
							rssi.length, i, false);

					// retry native returned with result
					while (err == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						err = GenericWifiManagerNatives.scanAPrssiAt(rssi, 0,
								rssi.length, i, true);
					}

					// check errors
					if (err < 0) {
						lastError = err;
						continue;
					}

					/* Get Security mode */
					int curAPSecurityMode = GenericWifiManagerNatives.scanAPsecurityModeAt(i, false);

					// retry native returned with result
					while (curAPSecurityMode == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						curAPSecurityMode = GenericWifiManagerNatives.scanAPsecurityModeAt(i, true);
					}

					// check errors
					if (curAPSecurityMode < 0) {
						lastError = curAPSecurityMode;
						continue;
					}

					/* Get Wps mode */
					int curAPWPSMode = GenericWifiManagerNatives.scanAPwpsModeAt(i, false);

					// retry native returned with result
					while (curAPWPSMode == J_NATIVE_CODE_BLOCKED_WITHOUT_RESULT) {
						curAPWPSMode = GenericWifiManagerNatives.scanAPwpsModeAt(i, true);
					}

					// check errors
					if (curAPWPSMode < 0) {
						lastError = curAPWPSMode;
						continue;
					}

					String curAPSSID = SNI.toJavaString(ssid);
					byte[] curAPBSSID = bssid.clone();
					AccessPoint curAP = new AccessPoint(curAPSSID, curAPBSSID, rssi[0], curAPChannel, SecurityModeNativeConstants.convertNativeSecurityModeToJavaMode(curAPSecurityMode), WPSModeNativeConstants.convertNativeWPSModeToJavaMode(curAPWPSMode));
					aps.add(curAP);
				}
			}
			if (lastError < 0 && aps.isEmpty()) {
				throw new IOException("I/O error (err: " + lastError + ")");
			}
		}
		AccessPoint[] apArray = new AccessPoint[aps.size()];
		return aps.toArray(apArray);
	}

	public AccessPoint[] scan() throws IOException {
		return scan(true);
	}


}