/*
 * Java
 *
 * Copyright 2016-2019 MicroEJ Corp. All rights reserved.
 * MicroEJ Corp. PROPRIETARY/CONFIDENTIAL. Use is subject to license terms.
 */
package ej.giml;

import java.io.IOException;
import java.lang.Thread.UncaughtExceptionHandler;
import java.util.NoSuchElementException;

import com.is2t.giml.runtime.nls.ImmutablesNLSFactory;
import com.is2t.giml.runtime.reader.GIMLReaderFactory;

import ej.components.dependencyinjection.ServiceLoaderFactory;
import ej.nls.NLS;

public class GIML {

	private final static String GIMLBinaryFileProperty = "giml.binary.file";
	private final static String GIMLNLSHeaderProperty = "gimlnls.header";
	private final static String GIMLNLSBinaryFileProperty = "gimlnls.binary.file";
	private final static String GIMLRuntimeVersionProperty = "giml.runtime.version";
	private final static String GIMLCompilerRuntimeVersionProperty = "giml.compiler.runtime.version";

	private static GIMLReader GIMLreader;
	private static GIMLFactoryProvider GIMLFactoryProvider;
	private static NLS nls;

	static {
		initialize();
	}

	/**
	 * Initializes the GIML components.
	 */
	private static void initialize() {

		// check GIML Runtime version
		checkGIMLRuntimeVersion();

		// get the GIML factory provider
		GIMLFactoryProvider = ServiceLoaderFactory.getServiceLoader().getService(GIMLFactoryProvider.class);
		if (GIMLFactoryProvider == null) {
			throw new NoSuchElementException("GIML Factory Provider not found");
		}

		// initialize the GIML reader
		try {
			GIMLReaderFactory readerFactory = ServiceLoaderFactory.getServiceLoader()
					.getService(GIMLReaderFactory.class);
			if (readerFactory == null) {
				throw new NoSuchElementException("GIMLReader Factory Provider not found");
			}
			String gimlBinaryPath = getGIMLBinaryPath(GIMLBinaryFileProperty);
			GIMLreader = readerFactory.newGIMLReader(gimlBinaryPath);

			ImmutablesNLSFactory immutablesNLSFactory = ServiceLoaderFactory.getServiceLoader()
					.getService(ImmutablesNLSFactory.class);
			if (immutablesNLSFactory == null) {
				throw new NoSuchElementException("GIMLImmutablesNLS Factory Provider not found");
			}

			String header = System.getProperty(GIMLNLSHeaderProperty);

			if (header != null) {
				nls = immutablesNLSFactory.newGIMLImmutablesNLS(header, getGIMLBinaryPath(GIMLNLSBinaryFileProperty));
			}
			GIMLreader.initialize(nls);

			// initialize the GIML factory provider
			GIMLFactoryProvider.initialize(GIMLreader);

		} catch (IOException e) {
			throw new RuntimeException("Could not initialize the reader", e);
		}
	}

	// check if the GIML runtime version supported by the compiler is compliant
	// with the current runtime version
	private static void checkGIMLRuntimeVersion() {
		int[] runtimeVersion = getVersion(GIMLRuntimeVersionProperty);
		int[] compilerRuntimeVersion = getVersion(GIMLCompilerRuntimeVersionProperty);
		int runtimeMajorVersion = runtimeVersion[0];
		int runtimeMinorVersion = runtimeVersion[1];
		int runtimePatchVersion = runtimeVersion[2];
		int compilerRuntimeMajorVersion = compilerRuntimeVersion[0];
		int compilerRuntimeMinorVersion = compilerRuntimeVersion[1];
		int compilerRuntimePatchVersion = compilerRuntimeVersion[2];
		if ((runtimeMajorVersion != compilerRuntimeMajorVersion) || (runtimeMinorVersion < compilerRuntimeMinorVersion)
				|| (runtimeMinorVersion == compilerRuntimeMinorVersion
						&& runtimePatchVersion < compilerRuntimePatchVersion)) {
			throw new RuntimeException("Incompatible GIML runtime version: " + runtimeMajorVersion + "."
					+ runtimeMinorVersion + "." + runtimePatchVersion + ". Supported version range ["
					+ compilerRuntimeMajorVersion + "." + compilerRuntimeMinorVersion + "."
					+ compilerRuntimePatchVersion + ", " + (compilerRuntimeMajorVersion + 1) + ".0.0[");
		}
	}

	// split the string value of the version defined by the given property and
	// return an array containing the major, the minor and the patch version
	private static int[] getVersion(String versionProperty) {
		String originalVersion = System.getProperty(versionProperty);
		if (originalVersion == null) {
			throw new RuntimeException("No GIML version defined with the property " + versionProperty);
		}
		String versionValue = originalVersion;
		try {
			int versionArray[] = new int[3];
			for (int i = -1; ++i < 3;) {
				if (i != 2) {
					int dotIndex = versionValue.indexOf(".");
					String versionPart = versionValue.substring(0, dotIndex);
					versionArray[i] = Integer.parseInt(versionPart);
					versionValue = versionValue.substring(dotIndex + 1);
				} else {
					int minusIndex = versionValue.indexOf("-");
					int plusIndex = versionValue.indexOf("+");
					if (minusIndex == -1 && plusIndex == -1) {
						versionArray[i] = Integer.parseInt(versionValue);
					} else if (minusIndex == -1) {
						versionArray[i] = Integer.parseInt(versionValue.substring(0, plusIndex));
					} else if (plusIndex == -1) {
						versionArray[i] = Integer.parseInt(versionValue.substring(0, minusIndex));
					} else {
						versionArray[i] = Integer.parseInt(versionValue.substring(0, Math.min(minusIndex, plusIndex)));
					}
				}
			}
			return versionArray;
		} catch (Exception e) {
			throw new RuntimeException(
					"Malformed GIML version " + originalVersion + " defined with the property " + versionProperty, e);
		}
	}

	private static String getGIMLBinaryPath(String gimlBinaryProperty) throws IOException {
		String gimlBinaryPath = System.getProperty(gimlBinaryProperty);
		if (gimlBinaryPath == null) {
			throw new IOException("No binary file defined with the property " + gimlBinaryProperty);
		}
		return gimlBinaryPath;
	}

	/**
	 * Gets the GIML reader.
	 *
	 * @return the GIML reader.
	 */
	public static final GIMLReader getGIMLReader() {
		return GIMLreader;
	}

	/**
	 * Creates the GIML root object identified by the given identifier.
	 * <p>
	 *
	 * @param identifier
	 *            the identifier of the root object to create.
	 * @return the created GIML root object.
	 * @throws IllegalArgumentException
	 *             if no GIML root object exists with the given identifier.
	 */
	public static synchronized Object createGIMLRootObject(int identifier) {
		// jump to the GIML root object position
		GIMLreader.seek(identifier);
		// create the GIML Object
		return createGIMLObject();
	}

	/**
	 * Creates the GIML object at the current reader position.
	 *
	 * @return the created GIML object.
	 */
	public static synchronized Object createGIMLObject() {
		// read the factortId
		int facroryId = GIMLreader.readVarUInt();
		// create the GIML object
		return GIMLFactoryProvider.createGIMLObject(facroryId);
	}

	/**
	 * Executes GIML specific code identified by the given id.
	 *
	 * If the code execution fails, the exception thrown is passed to the current thread uncaught exception handler, or
	 * the global uncaught exception handler or printed to the standard output.
	 *
	 * @param codeId
	 *            the code identifier.
	 * @return the result of the execution on success, otherwise null on error
	 */
	public static Object executeGIMLCode(int codeId) {
		try {
			return GIMLFactoryProvider.executeGIMLCode(codeId);
		} catch (Throwable e) {
			// handle the exception and continue
			handleUncaughtException(e);
			return null;
		}
	}

	private static void handleUncaughtException(Throwable e) {
		Thread currentThread = Thread.currentThread();
		UncaughtExceptionHandler h = currentThread.getUncaughtExceptionHandler();
		if (h != null) {
			h.uncaughtException(currentThread, e);
		} else {
			h = Thread.getDefaultUncaughtExceptionHandler();
			if (h != null) {
				h.uncaughtException(currentThread, e);
			} else {
				e.printStackTrace();
			}
		}
	}

	/**
	 * Returns the NLS object or <code>null</code> if there is no NLS messages.
	 *
	 * @return the NLS object or <code>null</code> if there is no NLS messages.
	 */
	public static NLS getNLS() {
		return nls;
	}

}
