/*
 * Java
 *
 * Copyright 2008-2021 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.display;

import com.is2t.microbsp.microui.natives.NSystemDisplay;
import com.is2t.tools.ArrayTools;

import ej.annotation.Nullable;
import ej.bon.Constants;
import ej.microui.MicroUI;
import ej.microui.MicroUIException;
import ej.microui.MicroUIProperties;

/**
 * microUI-API
 */
public class Font {

	/**
	 * microUI-API
	 */
	public static final int STYLE_PLAIN = 0;

	/**
	 * microUI-API
	 */
	public static final int STYLE_BOLD = 1;

	/**
	 * microUI-API
	 */
	public static final int STYLE_ITALIC = 1 << 1;

	private static @Nullable Font[] Fonts; // use getFonts()

	/**
	 * Native font identifier. Some fields are extracted (inlined) because there are
	 * often used.
	 */
	private final byte[] nativeData;

	/**
	 * List of all fonts loaded at same time than current font. This object is
	 * unique for all fonts.
	 */
	private final String list;

	/**
	 * Start index of path of this font in the previous list
	 */
	private final char pathStartIndex;

	/**
	 * End index of path of this font in the previous list
	 */
	private final char pathEndIndex;

	private final char height;
	private final char baseline;
	private final int[] identifiers;
	private final byte style;

	private Font(byte[] nativeData, String list, int pathStartIndex, int pathEndIndex) {
		this.nativeData = nativeData;
		this.list = list;
		this.pathStartIndex = (char) pathStartIndex;
		this.pathEndIndex = (char) pathEndIndex;

		this.height = FontData.getHeight(nativeData);
		this.baseline = FontData.getBaseline(nativeData);
		this.identifiers = FontData.getIdentifiers(nativeData);

		int styleFlags = 0;
		if (FontData.isBold(nativeData)) {
			styleFlags |= STYLE_BOLD;
		}
		if (FontData.isItalic(nativeData)) {
			styleFlags |= STYLE_ITALIC;
		}
		this.style = (byte) styleFlags;
	}

	/**
	 * microUI
	 *
	 * @return microUI-API
	 */
	public int[] getIdentifiers() {
		// spec does not specify if we have to return a copy
		return this.identifiers;
	}

	/**
	 * microUI-API
	 *
	 * @param identifier
	 *            microUI-API
	 * @return microUI-API
	 */
	public boolean isIdentifierSupported(int identifier) {
		int[] identifiers = getIdentifiers();
		for (int i = identifiers.length; --i >= 0;) {
			if (identifiers[i] == identifier) {
				return true;
			}
		}
		return false;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getStyle() {
		return this.style & (STYLE_BOLD | STYLE_ITALIC | STYLE_PLAIN);
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public String getDescriptor() {
		int filenameStartIndex = this.list.lastIndexOf('/', this.pathEndIndex) + 1;
		int filenameEndIndex = this.list.lastIndexOf('.', this.pathEndIndex);
		return this.list.substring(filenameStartIndex, filenameEndIndex);
	}

	private boolean isSamePath(String path) {
		return this.list.regionMatches(this.pathStartIndex, path, 0, path.length());
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public boolean isPlain() {
		// the style is Plain if no BOLD and no Italic (see getStyle())
		return (this.style & (STYLE_BOLD | STYLE_ITALIC)) == 0;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */

	public boolean isBold() {
		return (this.style & STYLE_BOLD) != 0;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public boolean isItalic() {
		return (this.style & STYLE_ITALIC) != 0;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public boolean isMonospaced() {
		return FontData.isMonospaced(this.nativeData);
	}

	/**
	 * microUI-API
	 *
	 * @param ch
	 *            microUI-API
	 * @return microUI-API
	 */
	public int charWidth(char ch) {
		return getWidth(new char[] { ch }, 0, 1, null);
	}

	/**
	 * not in API
	 *
	 * Copy string characters into a new array. Useful when cannot use EDC internal
	 * API.
	 *
	 * @param str
	 *            the string to extract.
	 * @return a characters array.
	 */
	public static char[] getStringChars(String str) {
		int length = str.length();
		char[] chars = new char[length];
		str.getChars(0, length, chars, 0);
		return chars;
	}

	/**
	 * microUI-API
	 *
	 * @param str
	 *            microUI-API
	 * @return microUI-API
	 */
	public int stringWidth(String str) {
		assert str != null;
		int length = str.length(); // can throw NPE (spec)
		if (Constants.getBoolean(MicroUIProperties.EDC_INTERNAL)) {
			return getWidth(str.chars, str.offset, length, null);
		} else {
			return getWidth(getStringChars(str), 0, length, null);
		}

	}

	/**
	 * microUI-API
	 *
	 * @param str
	 *            microUI-API
	 * @param offset
	 *            microUI-API
	 * @param len
	 *            microUI-API
	 * @return microUI-API
	 */
	public int substringWidth(String str, int offset, int len) {
		assert str != null;
		try {
			// check bounds, throw ArrayIndexOutOfBoundsException if wrong
			// values; str.length() can throw NPE
			ArrayTools.checkArrayBounds(str.length(), offset, len);
		} catch (ArrayIndexOutOfBoundsException e) {
			throw new StringIndexOutOfBoundsException();
		}

		// perform this check AFTER checkArrayBounds (spec)

		if (Constants.getBoolean(MicroUIProperties.EDC_INTERNAL)) {
			return getWidth(str.chars, str.offset + offset, len, null);
		} else {
			return getWidth(getStringChars(str), offset, len, null);
		}
	}

	/**
	 * @return native width (ratio is not applied)
	 */
	/* default */ int getWidth(char[] ch, int offset, int length, @Nullable int[] indexes) {
		return Font.checkFontRuntimeError(NSystemDisplay.charsWidth(getSNIContext(), ch, offset, length, indexes));
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public static Font getDefaultFont() {
		MicroUI.checkRunning();
		try {
			return getFontsInternal()[0];
		} catch (Exception e) {
			throw new MicroUIException(MicroUIException.NO_FONT);
		}
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public static Font[] getAllFonts() {
		// copy the array of fonts
		Font[] fonts = getFonts();
		int nbFonts = fonts.length;
		Font[] copy = new Font[nbFonts];
		System.arraycopy(fonts, 0, copy, 0, nbFonts);
		return copy;
	}

	/**
	 * microUI-API
	 *
	 * @param path
	 *            microUI-API
	 *
	 * @return microUI-API
	 */
	public static Font getFont(String path) {
		assert path != null;
		String p = MicroUIException.checkResourcePath(path);
		Font[] fonts = getFonts();
		for (Font font : fonts) {
			if (font.isSamePath(p)) {
				return font;
			}
		}
		throw new MicroUIException(MicroUIException.RESOURCE_INVALID_PATH, p);
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getHeight() {
		return this.height;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public int getBaselinePosition() {
		return this.baseline;
	}

	/**
	 * microUI-API
	 *
	 * @return microUI-API
	 */
	public byte[] getSNIContext() {
		return this.nativeData;
	}

	private static synchronized Font[] getFonts() {
		MicroUI.checkRunning();
		return getFontsInternal();
	}

	/**
	 * Lazy initialization of Fonts.
	 * <p>
	 * Synchronization must be ensured by the caller.
	 */
	/* default */ static Font[] getFontsInternal() {
		if (Fonts == null) {
			// No injection - default case
			Font.Fonts = initializeAllFonts();
		}
		return Fonts;
	}

	/* default */ static Font[] initializeAllFonts() {
		MicroUI.checkRunning();
		return addFonts(Constants.getString(MicroUIProperties.FONTS_LIST), new Font[0]);
	}

	private static void checkSecurityManager() {
		if (Constants.getBoolean(MicroUIProperties.CONSTANT_USE_SECURITYMANAGER)) {
			SecurityManager sm = System.getSecurityManager();
			if (sm != null) {
				sm.checkPermission(new FontPermission());
			}
		}
	}

	/* default */ static Font[] addFonts(final String list, Font[] fonts) {
		// the try-catch block should be written inside the if-constants block. But
		// the issue M0081MEJA-842 prevents it.
		// 96 bytes in flash can be won moving this try-catch.
		try {
			checkSecurityManager();
		} catch (SecurityException e) {
			// Not allowed to add font - silently complete
			return fonts;
		}

		// add all user fonts
		int index = 0;
		int listLength = list.length();
		while (index < listLength) {
			// get the next index of ':'
			int sepIndex = list.indexOf(MicroUIProperties.FONTS_LIST_SEPARATOR, index);

			// it was the last identifier
			if (sepIndex == -1) {
				sepIndex = listLength;
			}

			// add all MicroUI fonts for this path
			fonts = addFont(list, fonts, index, sepIndex);

			// update index
			index = sepIndex + 1;
		}

		return fonts;
	}

	private static Font[] addFont(final String list, Font[] microUIFonts, int pathStartIndex, int pathEndIndex) {
		// trim path manually
		while (pathStartIndex < pathEndIndex && list.charAt(pathStartIndex) <= ' ') {
			pathStartIndex++;
		}
		while (pathStartIndex < pathEndIndex && list.charAt(pathEndIndex - 1) <= ' ') {
			pathEndIndex--;
		}

		if (pathStartIndex >= pathEndIndex) {
			return microUIFonts; // empty
		}

		String path = list.substring(pathStartIndex, pathEndIndex);

		if (path.charAt(0) != '/') {
			MicroUI.MicroUI.errorLog(new MicroUIException(MicroUIException.RESOURCE_INVALID_PATH, path));
			return microUIFonts; // empty
		}

		try {
			// add NULL character at the end of string to be able load an external font
			String nativePath = path + '\0';

			// load font and retrieve some font characteristics
			byte[] nativeFontData = FontData.allocate();
			checkFontLoadingError(
					NSystemDisplay.addInternalFont(nativePath.getBytes(), nativePath.length(), nativeFontData));

			// create a font object for the ejf file
			Font font = new Font(nativeFontData, list, pathStartIndex, pathEndIndex);

			microUIFonts = (Font[]) ArrayTools.add(microUIFonts, font);
		} catch (Exception e) {
			// just print a warning and try to load remaining fonts
			MicroUI.MicroUI.errorLog(e);
		}

		return microUIFonts;
	}

	/**
	 * Check the error returned during a font use: drawText etc. Does not check the
	 * font loading error.
	 *
	 * @param ret
	 *            the potential error or the original return code
	 * @return the original return code when there is no error
	 * @throws MicroUIException
	 *             when an error is detected
	 */
	private static int checkFontRuntimeError(int ret) {
		/*
		 * WARNING this code is duplicated in microui-extended
		 */

		if (ret == NSystemDisplay.INVALID_FILE) {
			// error when reading the font
			throw new MicroUIException(MicroUIException.RESOURCE_INVALID_FILE);
		} else if (ret == NSystemDisplay.FILE_CLOSED) {
			// font (master font data) has been closed
			throw new MicroUIException(MicroUIException.RESOURCE_CLOSED);
		}
		return ret;
	}

	/**
	 * Check the error returned during a font loading: addFont, getIdentifiers etc.
	 *
	 * @param ret
	 *
	 * @return
	 */
	/* default */ static int checkFontLoadingError(int ret) {
		if (ret == NSystemDisplay.INVALID_PATH) {
			throw new MicroUIException(MicroUIException.RESOURCE_INVALID_PATH);
		} else if (ret == NSystemDisplay.INVALID_FILE) {
			throw new MicroUIException(MicroUIException.RESOURCE_INVALID_FILE);
		}
		return ret;
	}

	private int getDrawingWidth(GraphicsContext gc, char[] data, int offset, int length, int x) {
		// Width is useless in nominal case:
		// - ellipsis is false
		// - x start coordinate is in the clip
		// In this case, no need to retrieve width: it is never used by native algorithm
		//
		// Exceptions:
		// - If ellipsis is true, native algorithm requires the string width to
		// know when drawing the ellipsis.
		// - If x start coordinate is out of the clip on the left, native
		// algorithm requires the string width to know if the full string is
		// outside the clip on the left. In this case, no need to draw the
		// string.
		return gc.getEllipsisWidth() == 0 && x >= gc.getClipX() /* isOutsideClipOnTheLeft */ ? 0
				: getWidth(data, offset, length, null);
	}

	/* default */ void drawChars(GraphicsContext gc, char[] data, int offset, int length, int x, int y) {
		drawChars(gc, data, offset, length, x, y, null, getDrawingWidth(gc, data, offset, length, x));
	}

	/**
	 * not in API
	 *
	 * @param width
	 *            size of string in pixel. 0 is a valid value
	 * @see #getDrawingWidth(GraphicsContext, char[], int, int, int)
	 */
	/* default */ void drawChars(GraphicsContext gc, char[] data, int offset, int length, int x, int y,
			@Nullable int[] indexes, int width) {
		Font.checkFontRuntimeError(NSystemDisplay.drawChars(gc.getSNIContext(), getSNIContext(),
				x + gc.getTranslationX(), y + gc.getTranslationY(), data, offset, length, width, indexes));
	}

}
