/*
 * Java
 *
 * Copyright 2012-2020 MicroEJ Corp. All rights reserved.
 * This library is provided in source code for use, modification and test, subject to license terms.
 * Any modification of the source code will break MicroEJ Corp. warranties on the whole library.
 */
package ej.microui.event.generator;

import ej.microui.MicroUIPump;
import ej.microui.event.EventGenerator;
import ej.microui.event.EventHandler;

/**
 * Generic event generator.
 * <p>
 * Generic communication to receive events.
 */
public abstract class GenericEventGenerator extends EventGenerator {

	// masks & shifts
	private static final int DRIVER_LENGTH_SHIFT = 0;
	private static final int DRIVER_LENGTH_MASK = 0xff << DRIVER_LENGTH_SHIFT;

	@Override
	public final void convert(MicroUIPump pump, int inputEvent, EventHandler eventHandler) throws InterruptedException {
		// get the number of custom events
		int nbEvents = (inputEvent & DRIVER_LENGTH_MASK) >>> DRIVER_LENGTH_SHIFT;
		if (nbEvents == 1) {
			// only one event
			eventReceived(pump.read());
		} else {
			// several events
			int[] events = new int[nbEvents];

			// read all events
			for (int i = -1; ++i < nbEvents;) { // respect the order
				events[i] = pump.read();
			}

			eventsReceived(events);
		}
	}

	/**
	 * microUI-API Called at startup to configure the event generator with the
	 * specific properties set in the .microui file.
	 *
	 * @param name
	 *            the property name
	 * @param value
	 *            the property value
	 */
	public abstract void setProperty(String name, String value);

	/**
	 * microUI-API Called by MicroUI framework when a custom event has been received
	 * from native world. The current custom event contains only one 32-bit value.
	 *
	 * @param event
	 *            the 32-bit custom event value.
	 */
	protected abstract void eventReceived(int event);

	/**
	 * microUI-API Called by MicroUI framework when a custom event has been received
	 * from native world. The current custom event contains several 32-bit values.
	 *
	 * @param events
	 *            the 32-bit custom event values.
	 */
	protected abstract void eventsReceived(int[] events);
}
