/*
 * Copyright (c) 1997, 2013, Oracle and/or its affiliates. All rights reserved.
 * Copyright (C) 2015-2018, IS2T - EDC compliance and optimizations.
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This code is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 2 only, as
 * published by the Free Software Foundation.  Oracle designates this
 * particular file as subject to the "Classpath" exception as provided
 * by Oracle in the LICENSE file that accompanied this code.
 *
 * This code is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * version 2 for more details (a copy is included in the LICENSE file that
 * accompanied this code).
 *
 * You should have received a copy of the GNU General Public License version
 * 2 along with this work; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 * Please contact Oracle, 500 Oracle Parkway, Redwood Shores, CA 94065 USA
 * or visit www.oracle.com if you need additional information or have any
 * questions.
 */

package java.security.cert;

import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.SignatureException;

import com.microej.security.util.Arrays;

public abstract class Certificate implements java.io.Serializable {

	private static final long serialVersionUID = -3585440601605666277L;

	// the certificate type
	private final String type;

	/** Cache the hash code for the certiticate */
	private int hash = -1; // Default to -1

	protected Certificate(String type) {
		this.type = type;
	}

	public final String getType() {
		return this.type;
	}

	/**
	 * Compares this certificate for equality with the specified object. If the {@code other} object is an
	 * {@code instanceof} {@code Certificate}, then its encoded form is retrieved and compared with the encoded form of
	 * this certificate.
	 *
	 * @param other
	 *            the object to test for equality with this certificate.
	 * @return true iff the encoded forms of the two certificates match, false otherwise.
	 */
	@Override
	public boolean equals(Object other) {
		if (this == other) {
			return true;
		}
		if (!(other instanceof Certificate)) {
			return false;
		}
		try {
			byte[] thisCert = getEncoded();
			byte[] otherCert = ((Certificate) other).getEncoded();

			return Arrays.equals(thisCert, otherCert);
		} catch (CertificateException e) {
			return false;
		}
	}

	/**
	 * Returns a hashcode value for this certificate from its encoded form.
	 *
	 * @return the hashcode value.
	 */
	@Override
	public int hashCode() {
		int h = hash;
		if (h == -1) {
			try {
				h = Arrays.hashCode(getEncoded());
			} catch (CertificateException e) {
				h = 0;
			}
			hash = h;
		}
		return h;
	}

	/**
	 * Returns the encoded form of this certificate. It is assumed that each certificate type would have only a single
	 * form of encoding; for example, X.509 certificates would be encoded as ASN.1 DER.
	 *
	 * @return the encoded form of this certificate
	 *
	 * @exception CertificateEncodingException
	 *                if an encoding error occurs.
	 */
	public abstract byte[] getEncoded() throws CertificateEncodingException;

	/**
	 * Verifies that this certificate was signed using the private key that corresponds to the specified public key.
	 *
	 * @param key
	 *            the PublicKey used to carry out the verification.
	 *
	 * @exception NoSuchAlgorithmException
	 *                on unsupported signature algorithms.
	 * @exception InvalidKeyException
	 *                on incorrect key.
	 * @exception NoSuchProviderException
	 *                if there's no default provider.
	 * @exception SignatureException
	 *                on signature errors.
	 * @exception CertificateException
	 *                on encoding errors.
	 */
	public abstract void verify(PublicKey key) throws CertificateException, NoSuchAlgorithmException,
	InvalidKeyException, NoSuchProviderException, SignatureException;

	/**
	 * Verifies that this certificate was signed using the private key that corresponds to the specified public key.
	 * This method uses the signature verification engine supplied by the specified provider.
	 *
	 * @param key
	 *            the PublicKey used to carry out the verification.
	 * @param sigProvider
	 *            the name of the signature provider.
	 *
	 * @exception NoSuchAlgorithmException
	 *                on unsupported signature algorithms.
	 * @exception InvalidKeyException
	 *                on incorrect key.
	 * @exception NoSuchProviderException
	 *                on incorrect provider.
	 * @exception SignatureException
	 *                on signature errors.
	 * @exception CertificateException
	 *                on encoding errors.
	 */
	public abstract void verify(PublicKey key, String sigProvider) throws CertificateException,
	NoSuchAlgorithmException, InvalidKeyException, NoSuchProviderException, SignatureException;

	/**
	 * Returns a string representation of this certificate.
	 *
	 * @return a string representation of this certificate.
	 */
	@Override
	public abstract String toString();

	/**
	 * Gets the public key from this certificate.
	 *
	 * @return the public key.
	 */
	public abstract PublicKey getPublicKey();
}
